"""

Parameters X:
    Cells:           Single compartment, conductance based, passive cell
    Chem Synapses:   Analogue/graded synapses; continuous transmission (voltage dependent)
    Gap junctions:   Electrical connection; current linerly depends on difference in voltages

ASSESSMENT:
    TBC...

"""

from neuroml import Cell
from neuroml import Morphology
from neuroml import Point3DWithDiam
from neuroml import Segment
from neuroml import BiophysicalProperties
from neuroml import IntracellularProperties
from neuroml import Resistivity
from neuroml import MembraneProperties
from neuroml import InitMembPotential
from neuroml import SpecificCapacitance
from neuroml import ChannelDensity
from neuroml import SpikeThresh

from neuroml import GapJunction
from neuroml import PulseGenerator

from c302.bioparameters import c302ModelPrototype


class ParameterisedModel(c302ModelPrototype):
    def __init__(self):
        super(ParameterisedModel, self).__init__()
        self.level = "X"
        self.custom_component_types_definitions = ["cell_X.xml", "custom_synapses.xml"]

        self.set_default_bioparameters()
        self.print_("Set default parameters for %s" % self.level)

    def set_default_bioparameters(self):
        self.add_bioparameter("cell_diameter", "5.641895835477563", "BlindGuess", "0.1")
        self.add_bioparameter("muscle_length", "5.641895835477563", "BlindGuess", "0.1")

        self.add_bioparameter("initial_memb_pot", "-40 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "specific_capacitance", "1 uF_per_cm2", "BlindGuess", "0.1"
        )

        self.add_bioparameter("muscle_spike_thresh", "-20 mV", "BlindGuess", "0.1")
        self.add_bioparameter("neuron_spike_thresh", "-20 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "muscle_leak_cond_density", "0.01 mS_per_cm2", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_leak_cond_density", "0.01 mS_per_cm2", "BlindGuess", "0.1"
        )

        self.add_bioparameter("leak_erev", "-35 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "neuron_to_neuron_exc_syn_conductance", "100 pS", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_to_muscle_exc_syn_conductance", "0.00 pS", "BlindGuess", "0.1"
        )

        self.add_bioparameter("exc_syn_ar", "1 per_s", "BlindGuess", "0.1")
        self.add_bioparameter("exc_syn_ad", "5 per_s", "BlindGuess", "0.1")
        self.add_bioparameter("exc_syn_beta", "0.125 per_mV", "BlindGuess", "0.1")
        self.add_bioparameter("exc_syn_vth", "0 mV", "BlindGuess", "0.1")
        self.add_bioparameter("exc_syn_erev", "0 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "neuron_to_neuron_inh_syn_conductance", "100 pS", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_to_muscle_inh_syn_conductance", "0.00 nS", "BlindGuess", "0.1"
        )

        self.add_bioparameter("inh_syn_ar", "1 per_s", "BlindGuess", "0.1")
        self.add_bioparameter("inh_syn_ad", "5 per_s", "BlindGuess", "0.1")
        self.add_bioparameter("inh_syn_beta", "0.125 per_mV", "BlindGuess", "0.1")
        self.add_bioparameter("inh_syn_vth", "0 mV", "BlindGuess", "0.1")
        self.add_bioparameter("inh_syn_erev", "-45 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "neuron_to_neuron_elec_syn_gbase", "100 pS", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_to_muscle_elec_syn_gbase", "100 pS", "BlindGuess", "0.1"
        )

        self.add_bioparameter(
            "unphysiological_offset_current", "0 pA", "KnownError", "0"
        )  # Can be activated later
        self.add_bioparameter(
            "unphysiological_offset_current_del", "0 ms", "KnownError", "0"
        )
        self.add_bioparameter(
            "unphysiological_offset_current_dur", "2000 ms", "KnownError", "0"
        )

    def create_models(self):
        self.create_generic_muscle_cell()
        self.create_generic_neuron_cell()
        self.create_offsetcurrent()
        self.create_neuron_to_neuron_syn()
        self.create_neuron_to_muscle_syn()

    def create_generic_muscle_cell(self):
        self.generic_muscle_cell = Cell(id="GenericMuscleCell")

        morphology = Morphology()
        morphology.id = "morphology_" + self.generic_muscle_cell.id

        self.generic_muscle_cell.morphology = morphology

        prox_point = Point3DWithDiam(
            x="0", y="0", z="0", diameter=self.get_bioparameter("cell_diameter").value
        )
        dist_point = Point3DWithDiam(
            x="0",
            y=self.get_bioparameter("muscle_length").value,
            z="0",
            diameter=self.get_bioparameter("cell_diameter").value,
        )

        segment = Segment(id="0", name="soma", proximal=prox_point, distal=dist_point)

        morphology.segments.append(segment)

        self.generic_muscle_cell.biophysical_properties = BiophysicalProperties(
            id="biophys_" + self.generic_muscle_cell.id
        )

        mp = MembraneProperties()
        self.generic_muscle_cell.biophysical_properties.membrane_properties = mp

        mp.init_memb_potentials.append(
            InitMembPotential(value=self.get_bioparameter("initial_memb_pot").value)
        )

        mp.specific_capacitances.append(
            SpecificCapacitance(
                value=self.get_bioparameter("specific_capacitance").value
            )
        )

        mp.spike_threshes.append(
            SpikeThresh(value=self.get_bioparameter("muscle_spike_thresh").value)
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter("muscle_leak_cond_density").value,
                id="Leak_all",
                ion_channel="Leak",
                erev=self.get_bioparameter("leak_erev").value,
                ion="non_specific",
            )
        )

        ip = IntracellularProperties()
        self.generic_muscle_cell.biophysical_properties.intracellular_properties = ip

        # NOTE: resistivity/axial resistance not used for single compartment cell models, so value irrelevant!
        ip.resistivities.append(Resistivity(value="0.1 kohm_cm"))

    def create_generic_neuron_cell(self):
        self.generic_neuron_cell = Cell(id="GenericNeuronCell")

        morphology = Morphology()
        morphology.id = "morphology_" + self.generic_neuron_cell.id

        self.generic_neuron_cell.morphology = morphology

        prox_point = Point3DWithDiam(
            x="0", y="0", z="0", diameter=self.get_bioparameter("cell_diameter").value
        )
        dist_point = Point3DWithDiam(
            x="0", y="0", z="0", diameter=self.get_bioparameter("cell_diameter").value
        )

        segment = Segment(id="0", name="soma", proximal=prox_point, distal=dist_point)

        morphology.segments.append(segment)

        self.generic_neuron_cell.biophysical_properties = BiophysicalProperties(
            id="biophys_" + self.generic_neuron_cell.id
        )

        mp = MembraneProperties()
        self.generic_neuron_cell.biophysical_properties.membrane_properties = mp

        mp.init_memb_potentials.append(
            InitMembPotential(value=self.get_bioparameter("initial_memb_pot").value)
        )

        mp.specific_capacitances.append(
            SpecificCapacitance(
                value=self.get_bioparameter("specific_capacitance").value
            )
        )

        mp.spike_threshes.append(
            SpikeThresh(value=self.get_bioparameter("neuron_spike_thresh").value)
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter("neuron_leak_cond_density").value,
                id="Leak_all",
                ion_channel="Leak",
                erev=self.get_bioparameter("leak_erev").value,
                ion="non_specific",
            )
        )

        ip = IntracellularProperties()
        self.generic_neuron_cell.biophysical_properties.intracellular_properties = ip

        # NOTE: resistivity/axial resistance not used for single compartment cell models, so value irrelevant!
        ip.resistivities.append(Resistivity(value="0.1 kohm_cm"))

    def create_offsetcurrent(self):
        self.offset_current = PulseGenerator(
            id="offset_current",
            delay=self.get_bioparameter("unphysiological_offset_current_del").value,
            duration=self.get_bioparameter("unphysiological_offset_current_dur").value,
            amplitude=self.get_bioparameter("unphysiological_offset_current").value,
        )

    def create_neuron_to_neuron_syn(self):
        self.neuron_to_neuron_exc_syn = GradedSynapse2(
            id="neuron_to_neuron_exc_syn",
            conductance=self.get_bioparameter(
                "neuron_to_neuron_exc_syn_conductance"
            ).value,
            ar=self.get_bioparameter("exc_syn_ar").value,
            ad=self.get_bioparameter("exc_syn_ad").value,
            beta=self.get_bioparameter("exc_syn_beta").value,
            vth=self.get_bioparameter("exc_syn_vth").value,
            erev=self.get_bioparameter("exc_syn_erev").value,
        )

        self.neuron_to_neuron_inh_syn = GradedSynapse2(
            id="neuron_to_neuron_inh_syn",
            conductance=self.get_bioparameter(
                "neuron_to_neuron_inh_syn_conductance"
            ).value,
            ar=self.get_bioparameter("inh_syn_ar").value,
            ad=self.get_bioparameter("inh_syn_ad").value,
            beta=self.get_bioparameter("inh_syn_beta").value,
            vth=self.get_bioparameter("inh_syn_vth").value,
            erev=self.get_bioparameter("inh_syn_erev").value,
        )

        self.neuron_to_neuron_elec_syn = GapJunction(
            id="neuron_to_neuron_elec_syn",
            conductance=self.get_bioparameter("neuron_to_neuron_elec_syn_gbase").value,
        )

    def create_neuron_to_muscle_syn(self):
        self.neuron_to_muscle_exc_syn = GradedSynapse2(
            id="neuron_to_muscle_exc_syn",
            conductance=self.get_bioparameter(
                "neuron_to_muscle_exc_syn_conductance"
            ).value,
            ar=self.get_bioparameter("exc_syn_ar").value,
            ad=self.get_bioparameter("exc_syn_ad").value,
            beta=self.get_bioparameter("exc_syn_beta").value,
            vth=self.get_bioparameter("exc_syn_vth").value,
            erev=self.get_bioparameter("exc_syn_erev").value,
        )

        self.neuron_to_muscle_inh_syn = GradedSynapse2(
            id="neuron_to_muscle_inh_syn",
            conductance=self.get_bioparameter(
                "neuron_to_muscle_inh_syn_conductance"
            ).value,
            ar=self.get_bioparameter("inh_syn_ar").value,
            ad=self.get_bioparameter("inh_syn_ad").value,
            beta=self.get_bioparameter("inh_syn_beta").value,
            vth=self.get_bioparameter("inh_syn_vth").value,
            erev=self.get_bioparameter("inh_syn_erev").value,
        )

        self.neuron_to_muscle_elec_syn = GapJunction(
            id="neuron_to_muscle_elec_syn",
            conductance=self.get_bioparameter("neuron_to_muscle_elec_syn_gbase").value,
        )

    def get_elec_syn(self, pre_cell, post_cell, type):
        self.found_specific_param = False
        if type == "neuron_to_neuron":
            gbase = self.get_conn_param(
                pre_cell,
                post_cell,
                "%s_to_%s_elec_syn_%s",
                "neuron_to_neuron_elec_syn_%s",
                "gbase",
            )
            conn_id = "neuron_to_neuron_elec_syn"
        elif type == "neuron_to_muscle":
            gbase = self.get_conn_param(
                pre_cell,
                post_cell,
                "%s_to_%s_elec_syn_%s",
                "neuron_to_muscle_elec_syn_%s",
                "gbase",
            )
            conn_id = "neuron_to_muscle_elec_syn"

        if self.found_specific_param:
            conn_id = "%s_to_%s_elec_syn" % (pre_cell, post_cell)

        return GapJunction(id=conn_id, conductance=gbase)

    def get_exc_syn(self, pre_cell, post_cell, type):
        self.found_specific_param = False

        specific_param_template = "%s_to_%s_exc_syn_%s"
        if type == "neuron_to_neuron":
            conductance = self.get_conn_param(
                pre_cell,
                post_cell,
                specific_param_template,
                "neuron_to_neuron_exc_syn_%s",
                "conductance",
            )
            erev = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "erev"
            )
            ar = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "ar"
            )
            ad = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "ad"
            )
            beta = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "beta"
            )
            vth = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "vth"
            )

            conn_id = "neuron_to_neuron_exc_syn"

        elif type == "neuron_to_muscle":
            conductance = self.get_conn_param(
                pre_cell,
                post_cell,
                specific_param_template,
                "neuron_to_muscle_exc_syn_%s",
                "conductance",
            )
            erev = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "erev"
            )
            ar = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "ar"
            )
            ad = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "ad"
            )
            beta = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "beta"
            )
            vth = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "exc_syn_%s", "vth"
            )
            conn_id = "neuron_to_muscle_exc_syn"

        if self.found_specific_param:
            conn_id = "%s_to_%s_exc_syn" % (pre_cell, post_cell)

        return GradedSynapse2(
            id=conn_id,
            conductance=conductance,
            ar=ar,
            ad=ad,
            beta=beta,
            vth=vth,
            erev=erev,
        )

    def get_inh_syn(self, pre_cell, post_cell, type):
        self.found_specific_param = False

        specific_param_template = "%s_to_%s_inh_syn_%s"
        if type == "neuron_to_neuron":
            conductance = self.get_conn_param(
                pre_cell,
                post_cell,
                specific_param_template,
                "neuron_to_neuron_inh_syn_%s",
                "conductance",
            )
            erev = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "erev"
            )
            ar = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "ar"
            )
            ad = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "ad"
            )
            beta = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "beta"
            )
            vth = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "vth"
            )

            conn_id = "neuron_to_neuron_inh_syn"

        elif type == "neuron_to_muscle":
            conductance = self.get_conn_param(
                pre_cell,
                post_cell,
                specific_param_template,
                "neuron_to_muscle_inh_syn_%s",
                "conductance",
            )
            erev = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "erev"
            )
            ar = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "ar"
            )
            ad = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "ad"
            )
            beta = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "beta"
            )
            vth = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "inh_syn_%s", "vth"
            )

            conn_id = "neuron_to_muscle_inh_syn"

        if self.found_specific_param:
            conn_id = "%s_to_%s_inh_syn" % (pre_cell, post_cell)

        return GradedSynapse2(
            id=conn_id,
            conductance=conductance,
            ar=ar,
            ad=ad,
            beta=beta,
            vth=vth,
            erev=erev,
        )

    def create_n_connection_synapse(self, prototype_syn, n, nml_doc, existing_synapses):
        if prototype_syn.id in existing_synapses:
            return existing_synapses[prototype_syn.id]

        if isinstance(prototype_syn, GradedSynapse2):
            existing_synapses[prototype_syn.id] = prototype_syn
            nml_doc.graded_synapses.append(prototype_syn)
            return prototype_syn
        else:
            return super(ParameterisedModel, self).create_n_connection_synapse(
                prototype_syn, n, nml_doc, existing_synapses
            )

    def is_analog_conn(self, syn):
        return super(ParameterisedModel, self).is_analog_conn(syn) or isinstance(
            syn, GradedSynapse2
        )


class GradedSynapse2:
    def __init__(self, id, conductance, ar, ad, beta, vth, erev):
        self.id = id
        self.conductance = conductance
        self.ar = ar
        self.ad = ad
        self.beta = beta
        self.vth = vth
        self.erev = erev

    def export(self, outfile, level, namespace, name_, pretty_print=True, **kwargs_):
        outfile.write(
            "    " * level
            + '<gradedSynapse2 id="%s" conductance="%s" ar="%s" ad="%s" beta="%s" vth="%s" erev="%s"/>\n'
            % (
                self.id,
                self.conductance,
                self.ar,
                self.ad,
                self.beta,
                self.vth,
                self.erev,
            )
        )



# -*- coding: utf-8 -*-

############################################################

#    Utilities for reading/writing/parsing NeuroML 2 files

############################################################


from c302.ConnectomeReader import analyse_connections


def getSegmentIds(cell):
    seg_ids = []
    for segment in cell.morphology.segments:
        seg_ids.append(segment.id)

    return seg_ids


def get3DPosition(cell, segment_index, fraction_along):
    seg = cell.morphology.segments[segment_index]

    end = seg.distal

    start = seg.proximal
    if start is None:
        segs = getSegmentIds(cell)
        seg_index_parent = segs.index(seg.parent.segments)
        start = cell.morphology.segments[seg_index_parent].distal

    fx = fract(start.x, end.x, fraction_along)
    fy = fract(start.y, end.y, fraction_along)
    fz = fract(start.z, end.z, fraction_along)

    # print "(%f, %f, %f) is %f between (%f, %f, %f) and (%f, %f, %f)"%(fx,fy,fz,fraction_along,start.x,start.y,start.z,end.x,end.y,end.z)

    return fx, fy, fz


def fract(a, b, f):
    return a + (b - a) * f


if __name__ == "__main__":
    from WormNeuroAtlasReader import read_data, read_muscle_data

    cells, neuron_conns = read_data(include_nonconnected_cells=True)
    neurons2muscles, muscles, muscle_conns = read_muscle_data()

    analyse_connections(cells, neuron_conns, neurons2muscles, muscles, muscle_conns)



"""

Parameters D:
    Cells:           Multicompartmental, conductance based cell models with HH like ion channels
    Chem Synapses:   Event based, ohmic; one rise & one decay constant
    Gap junctions:   Electrical connection; current linerly depends on difference in voltages

ASSESSMENT:
    As with C, the use of event based synapses normally requires spiking in cells, so core neurons will
    have to have clear spikes.
    Also: either the cells have i) low internal resistance (cytoplasmic resistivity) & membrane potential changes rapidly propagate
    through cell or ii) high internal resistance and changes are concentrated around the soma. For i) this allows all synapses on
    cell (e.g. distant dendrites) to transmit if cell fires/depolarises, but it means bigger cells have much higher imput resistance,
    and so take much more syn input to respond.
    Note issue https://github.com/openworm/CElegansNeuroML/issues/71 regarding status of this

"""

from neuroml import Cell
from neuroml import Morphology
from neuroml import Point3DWithDiam
from neuroml import Segment
from neuroml import BiophysicalProperties
from neuroml import IntracellularProperties
from neuroml import Resistivity
from neuroml import Species
from neuroml import MembraneProperties
from neuroml import InitMembPotential
from neuroml import SpecificCapacitance
from neuroml import ChannelDensity
from neuroml import SpikeThresh
from neuroml import FixedFactorConcentrationModel

from neuroml import ExpTwoSynapse
from neuroml import GapJunction
from neuroml import PulseGenerator

from c302.bioparameters import c302ModelPrototype


class ParameterisedModel(c302ModelPrototype):
    def __init__(self):
        super(ParameterisedModel, self).__init__()
        self.level = "D"
        self.custom_component_types_definitions = "cell_C.xml"

        self.set_default_bioparameters()

    def set_default_bioparameters(self):
        self.add_bioparameter("cell_diameter", "5", "BlindGuess", "0.1")
        self.add_bioparameter("muscle_length", "20", "BlindGuess", "0.1")

        self.add_bioparameter("initial_memb_pot", "-45 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "specific_capacitance", "1 uF_per_cm2", "BlindGuess", "0.1"
        )

        self.add_bioparameter("resistivity", "12 kohm_cm", "BlindGuess", "0.1")

        self.add_bioparameter("muscle_spike_thresh", "-26 mV", "BlindGuess", "0.1")
        self.add_bioparameter("neuron_spike_thresh", "-26 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "muscle_leak_cond_density", "0.005 mS_per_cm2", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_leak_cond_density", "0.02 mS_per_cm2", "BlindGuess", "0.1"
        )
        self.add_bioparameter("leak_erev", "-50 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "muscle_k_slow_cond_density", "4 mS_per_cm2", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_k_slow_cond_density", "2 mS_per_cm2", "BlindGuess", "0.1"
        )
        self.add_bioparameter("k_slow_erev", "-60 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "muscle_k_fast_cond_density", "0.2 mS_per_cm2", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_k_fast_cond_density", "0.2 mS_per_cm2", "BlindGuess", "0.1"
        )
        self.add_bioparameter("k_fast_erev", "-60 mV", "BlindGuess", "0.1")

        self.add_bioparameter(
            "muscle_ca_boyle_cond_density", "2 mS_per_cm2", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_ca_boyle_cond_density", "2 mS_per_cm2", "BlindGuess", "0.1"
        )
        self.add_bioparameter("ca_boyle_erev", "40 mV", "BlindGuess", "0.1")

        self.add_bioparameter("ca_conc_decay_time", "11.5943 ms", "BlindGuess", "0.1")
        self.add_bioparameter(
            "ca_conc_rho", "0.000238919 mol_per_m_per_A_per_s", "BlindGuess", "0.1"
        )

        self.add_bioparameter(
            "neuron_to_neuron_chem_exc_syn_gbase", ".01 nS", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_to_muscle_chem_exc_syn_gbase", ".01 nS", "BlindGuess", "0.1"
        )

        self.add_bioparameter("chem_exc_syn_erev", "0 mV", "BlindGuess", "0.1")
        self.add_bioparameter("chem_exc_syn_rise", "1 ms", "BlindGuess", "0.1")
        self.add_bioparameter("chem_exc_syn_decay", "5 ms", "BlindGuess", "0.1")

        self.add_bioparameter(
            "neuron_to_neuron_chem_inh_syn_gbase", "3 nS", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_to_muscle_chem_inh_syn_gbase", "3 nS", "BlindGuess", "0.1"
        )

        self.add_bioparameter("chem_inh_syn_erev", "-60 mV", "BlindGuess", "0.1")
        self.add_bioparameter("chem_inh_syn_rise", "2 ms", "BlindGuess", "0.1")
        self.add_bioparameter("chem_inh_syn_decay", "40 ms", "BlindGuess", "0.1")

        self.add_bioparameter(
            "neuron_to_neuron_elec_syn_gbase", "0.0005 nS", "BlindGuess", "0.1"
        )
        self.add_bioparameter(
            "neuron_to_muscle_elec_syn_gbase", "0.0005 nS", "BlindGuess", "0.1"
        )

        self.add_bioparameter(
            "unphysiological_offset_current", "0 pA", "KnownError", "0"
        )  # Can be activated later
        self.add_bioparameter(
            "unphysiological_offset_current_del", "0 ms", "KnownError", "0"
        )
        self.add_bioparameter(
            "unphysiological_offset_current_dur", "2000 ms", "KnownError", "0"
        )

    def create_models(self):
        self.create_generic_muscle_cell()

        self.create_offsetcurrent_concentrationmodel()
        self.create_neuron_to_neuron_syn()
        self.create_neuron_to_muscle_syn()

    def create_generic_muscle_cell(self):
        self.generic_muscle_cell = Cell(id="GenericMuscleCell")

        morphology = Morphology()
        morphology.id = "morphology_" + self.generic_muscle_cell.id

        self.generic_muscle_cell.morphology = morphology

        prox_point = Point3DWithDiam(
            x="0", y="0", z="0", diameter=self.get_bioparameter("cell_diameter").value
        )
        dist_point = Point3DWithDiam(
            x="0",
            y=self.get_bioparameter("muscle_length").value,
            z="0",
            diameter=self.get_bioparameter("cell_diameter").value,
        )

        segment = Segment(id="0", name="soma", proximal=prox_point, distal=dist_point)

        morphology.segments.append(segment)

        self.generic_muscle_cell.biophysical_properties = BiophysicalProperties(
            id="biophys_" + self.generic_muscle_cell.id
        )

        mp = MembraneProperties()
        self.generic_muscle_cell.biophysical_properties.membrane_properties = mp

        mp.init_memb_potentials.append(
            InitMembPotential(value=self.get_bioparameter("initial_memb_pot").value)
        )

        mp.specific_capacitances.append(
            SpecificCapacitance(
                value=self.get_bioparameter("specific_capacitance").value
            )
        )

        mp.spike_threshes.append(
            SpikeThresh(value=self.get_bioparameter("muscle_spike_thresh").value)
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter("muscle_leak_cond_density").value,
                id="Leak_all",
                ion_channel="Leak",
                erev=self.get_bioparameter("leak_erev").value,
                ion="non_specific",
            )
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter("muscle_k_slow_cond_density").value,
                id="k_slow_all",
                ion_channel="k_slow",
                erev=self.get_bioparameter("k_slow_erev").value,
                ion="k",
            )
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter("muscle_k_fast_cond_density").value,
                id="k_fast_all",
                ion_channel="k_fast",
                erev=self.get_bioparameter("k_fast_erev").value,
                ion="k",
            )
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter(
                    "muscle_ca_boyle_cond_density"
                ).value,
                id="ca_boyle_all",
                ion_channel="ca_boyle",
                erev=self.get_bioparameter("ca_boyle_erev").value,
                ion="ca",
            )
        )

        ip = IntracellularProperties()
        self.generic_muscle_cell.biophysical_properties.intracellular_properties = ip

        # NOTE: resistivity/axial resistance not used for single compartment cell models, so value irrelevant!
        ip.resistivities.append(
            Resistivity(value=self.get_bioparameter("resistivity").value)
        )

        # NOTE: Ca reversal potential not calculated by Nernst, so initial_ext_concentration value irrelevant!
        species = Species(
            id="ca",
            ion="ca",
            concentration_model="CaPool",
            initial_concentration="0 mM",
            initial_ext_concentration="2E-6 mol_per_cm3",
        )

        ip.species.append(species)

    def create_neuron_cell(self, cell_name, morphology):
        cell = Cell(id=cell_name)

        cell.notes = "Cell model created by c302 with custom electrical parameters"

        cell.morphology = morphology

        cell.biophysical_properties = BiophysicalProperties(id="biophys_" + cell.id)

        mp = MembraneProperties()
        cell.biophysical_properties.membrane_properties = mp

        mp.init_memb_potentials.append(
            InitMembPotential(value=self.get_bioparameter("initial_memb_pot").value)
        )

        mp.specific_capacitances.append(
            SpecificCapacitance(
                value=self.get_bioparameter("specific_capacitance").value
            )
        )

        mp.spike_threshes.append(
            SpikeThresh(value=self.get_bioparameter("neuron_spike_thresh").value)
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter("neuron_leak_cond_density").value,
                id="Leak_all",
                ion_channel="Leak",
                erev=self.get_bioparameter("leak_erev").value,
                ion="non_specific",
            )
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter("neuron_k_slow_cond_density").value,
                id="k_slow_all",
                ion_channel="k_slow",
                erev=self.get_bioparameter("k_slow_erev").value,
                ion="k",
            )
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter("neuron_k_fast_cond_density").value,
                id="k_fast_all",
                ion_channel="k_fast",
                erev=self.get_bioparameter("k_fast_erev").value,
                ion="k",
            )
        )

        mp.channel_densities.append(
            ChannelDensity(
                cond_density=self.get_bioparameter(
                    "neuron_ca_boyle_cond_density"
                ).value,
                id="ca_boyle_all",
                ion_channel="ca_boyle",
                erev=self.get_bioparameter("ca_boyle_erev").value,
                ion="ca",
            )
        )

        ip = IntracellularProperties()
        cell.biophysical_properties.intracellular_properties = ip

        # NOTE: resistivity/axial resistance not used for single compartment cell models, so value irrelevant!
        ip.resistivities.append(
            Resistivity(value=self.get_bioparameter("resistivity").value)
        )

        # NOTE: Ca reversal potential not calculated by Nernst, so initial_ext_concentration value irrelevant!
        species = Species(
            id="ca",
            ion="ca",
            concentration_model="CaPool",
            initial_concentration="0 mM",
            initial_ext_concentration="2E-6 mol_per_cm3",
        )

        ip.species.append(species)

        return cell

    def create_offsetcurrent_concentrationmodel(self):
        self.offset_current = PulseGenerator(
            id="offset_current",
            delay=self.get_bioparameter("unphysiological_offset_current_del").value,
            duration=self.get_bioparameter("unphysiological_offset_current_dur").value,
            amplitude=self.get_bioparameter("unphysiological_offset_current").value,
        )

        self.concentration_model = FixedFactorConcentrationModel(
            id="CaPool",
            ion="ca",
            resting_conc="0 mM",
            decay_constant=self.get_bioparameter("ca_conc_decay_time").value,
            rho=self.get_bioparameter("ca_conc_rho").value,
        )

    def create_neuron_to_neuron_syn(self):
        self.neuron_to_neuron_exc_syn = ExpTwoSynapse(
            id="neuron_to_neuron_exc_syn",
            gbase=self.get_bioparameter("neuron_to_neuron_chem_exc_syn_gbase").value,
            erev=self.get_bioparameter("chem_exc_syn_erev").value,
            tau_decay=self.get_bioparameter("chem_exc_syn_decay").value,
            tau_rise=self.get_bioparameter("chem_exc_syn_rise").value,
        )

        self.neuron_to_neuron_inh_syn = ExpTwoSynapse(
            id="neuron_to_neuron_inh_syn",
            gbase=self.get_bioparameter("neuron_to_neuron_chem_inh_syn_gbase").value,
            erev=self.get_bioparameter("chem_inh_syn_erev").value,
            tau_decay=self.get_bioparameter("chem_inh_syn_decay").value,
            tau_rise=self.get_bioparameter("chem_inh_syn_rise").value,
        )

        self.neuron_to_neuron_elec_syn = GapJunction(
            id="neuron_to_neuron_elec_syn",
            conductance=self.get_bioparameter("neuron_to_neuron_elec_syn_gbase").value,
        )

    def create_neuron_to_muscle_syn(self):
        self.neuron_to_muscle_exc_syn = ExpTwoSynapse(
            id="neuron_to_muscle_exc_syn",
            gbase=self.get_bioparameter("neuron_to_muscle_chem_exc_syn_gbase").value,
            erev=self.get_bioparameter("chem_exc_syn_erev").value,
            tau_decay=self.get_bioparameter("chem_exc_syn_decay").value,
            tau_rise=self.get_bioparameter("chem_exc_syn_rise").value,
        )

        self.neuron_to_muscle_inh_syn = ExpTwoSynapse(
            id="neuron_to_muscle_inh_syn",
            gbase=self.get_bioparameter("neuron_to_muscle_chem_inh_syn_gbase").value,
            erev=self.get_bioparameter("chem_inh_syn_erev").value,
            tau_decay=self.get_bioparameter("chem_inh_syn_decay").value,
            tau_rise=self.get_bioparameter("chem_inh_syn_rise").value,
        )

        self.neuron_to_muscle_elec_syn = GapJunction(
            id="neuron_to_muscle_elec_syn",
            conductance=self.get_bioparameter("neuron_to_muscle_elec_syn_gbase").value,
        )

    def get_elec_syn(self, pre_cell, post_cell, type):
        self.found_specific_param = False
        if type == "neuron_to_neuron":
            gbase = self.get_conn_param(
                pre_cell,
                post_cell,
                "%s_to_%s_elec_syn_%s",
                "neuron_to_neuron_elec_syn_%s",
                "gbase",
            )
            conn_id = "neuron_to_neuron_elec_syn"
        elif type == "neuron_to_muscle":
            gbase = self.get_conn_param(
                pre_cell,
                post_cell,
                "%s_to_%s_elec_syn_%s",
                "neuron_to_muscle_elec_syn_%s",
                "gbase",
            )
            conn_id = "neuron_to_muscle_elec_syn"

        if self.found_specific_param:
            conn_id = "%s_to_%s_elec_syn" % (pre_cell, post_cell)

        return GapJunction(id=conn_id, conductance=gbase)

    def get_exc_syn(self, pre_cell, post_cell, type):
        self.found_specific_param = False

        specific_param_template = "%s_to_%s_chem_exc_syn_%s"
        if type == "neuron_to_neuron":
            gbase = self.get_conn_param(
                pre_cell,
                post_cell,
                specific_param_template,
                "neuron_to_neuron_chem_exc_syn_%s",
                "gbase",
            )
            erev = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_exc_syn_%s", "erev"
            )
            decay = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_exc_syn_%s", "decay"
            )
            rise = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_exc_syn_%s", "rise"
            )

            conn_id = "neuron_to_neuron_exc_syn"

        elif type == "neuron_to_muscle":
            gbase = self.get_conn_param(
                pre_cell,
                post_cell,
                specific_param_template,
                "neuron_to_muscle_chem_exc_syn_%s",
                "gbase",
            )
            erev = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_exc_syn_%s", "erev"
            )
            decay = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_exc_syn_%s", "decay"
            )
            rise = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_exc_syn_%s", "rise"
            )
            conn_id = "neuron_to_muscle_exc_syn"

        if self.found_specific_param:
            conn_id = "%s_to_%s_exc_syn" % (pre_cell, post_cell)

        return ExpTwoSynapse(
            id=conn_id, gbase=gbase, erev=erev, tau_decay=decay, tau_rise=rise
        )

    def get_inh_syn(self, pre_cell, post_cell, type):
        self.found_specific_param = False

        specific_param_template = "%s_to_%s_chem_inh_syn_%s"
        if type == "neuron_to_neuron":
            gbase = self.get_conn_param(
                pre_cell,
                post_cell,
                specific_param_template,
                "neuron_to_neuron_chem_inh_syn_%s",
                "gbase",
            )
            erev = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_inh_syn_%s", "erev"
            )
            decay = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_inh_syn_%s", "decay"
            )
            rise = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_inh_syn_%s", "rise"
            )

            conn_id = "neuron_to_neuron_inh_syn"

        elif type == "neuron_to_muscle":
            gbase = self.get_conn_param(
                pre_cell,
                post_cell,
                specific_param_template,
                "neuron_to_muscle_chem_inh_syn_%s",
                "gbase",
            )
            erev = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_inh_syn_%s", "erev"
            )
            decay = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_inh_syn_%s", "decay"
            )
            rise = self.get_conn_param(
                pre_cell, post_cell, specific_param_template, "chem_inh_syn_%s", "rise"
            )
            conn_id = "neuron_to_muscle_inh_syn"

        if self.found_specific_param:
            conn_id = "%s_to_%s_inh_syn" % (pre_cell, post_cell)

        return ExpTwoSynapse(
            id=conn_id, gbase=gbase, erev=erev, tau_decay=decay, tau_rise=rise
        )



"""

Tap-Withdrawal circuit still under development - it does not produce the correct behavior!

"""

import c302
import sys
import importlib

import neuroml.writers as writers

range_incl = lambda start, end: range(start, end + 1)


def setup(
    parameter_set,
    generate=False,
    duration=400,
    dt=0.05,
    target_directory="examples",
    data_reader=c302.FW_DATA_READER,
    param_overrides={},
    config_param_overrides={},
    verbose=True,
):
    ParameterisedModel = getattr(
        importlib.import_module("c302.parameters_%s" % parameter_set),
        "ParameterisedModel",
    )
    params = ParameterisedModel()

    params.set_bioparameter(
        "unphysiological_offset_current", "0pA", "Testing TapWithdrawal", "0"
    )
    params.set_bioparameter(
        "unphysiological_offset_current_del", "0 ms", "Testing TapWithdrawal", "0"
    )
    params.set_bioparameter(
        "unphysiological_offset_current_dur", "2000 ms", "Testing TapWithdrawal", "0"
    )

    VA_motors = ["VA%s" % c for c in range_incl(1, 12)]
    VB_motors = ["VB%s" % c for c in range_incl(1, 11)]
    DA_motors = ["DA%s" % c for c in range_incl(1, 9)]
    DB_motors = ["DB%s" % c for c in range_incl(1, 7)]
    DD_motors = ["DD%s" % c for c in range_incl(1, 6)]
    VD_motors = ["VD%s" % c for c in range_incl(1, 13)]
    AS_motors = ["AS%s" % c for c in range_incl(1, 11)]
    AS_motors = []
    TW_cells = [
        "AVAL",
        "AVAR",
        "AVBL",
        "AVBR",
        "PVCL",
        "PVCR",
        "AVDL",
        "AVDR",
        "DVA",
        "PVDL",
        "PVDR",
        "PLML",
        "PLMR",
        "AVM",
        "ALML",
        "ALMR",
    ]
    # TW_sensory = ["PLML", "PLMR", "AVM", "ALML", "ALMR"]
    all_motors = list(
        VA_motors
        + VB_motors
        + DA_motors
        + DB_motors
        + DD_motors
        + VD_motors
        + AS_motors
    )
    # all_motors = []

    # neurons = ['AVAL', 'AVAR', 'AVBL', 'AVBR', 'PVCL', 'PVCR', 'AVDL', 'AVDR', 'DVA', 'PVDL', 'PVDR', 'PLML', 'PLMR', 'AVM', 'ALML', 'ALMR']

    muscles_to_include = False
    # muscles_to_include = True # ALL muscles
    # muscles_to_include = ['MVL01', 'MVL10']

    cells = list(TW_cells + all_motors)

    cells_to_plot = list(cells)
    # cells_to_plot = ["VB5", "VA4"]

    cells += [  # 'AS1', 'AS10', 'AS11', 'AS2', 'AS3', 'AS4', 'AS5', 'AS6', 'AS7', 'AS8', 'AS9',
        # 'AVFL', 'AVFR', 'AVKR', 'AVL',
        # 'CEPVL', 'CEPVR',
        #
        #
        # 'DVB',
        # 'HSNL', 'HSNR',
        # 'IL1DL', 'IL1DR', 'IL1L', 'IL1R', 'IL1VL', 'IL1VR',
        # 'PDA', 'PDB',
        # 'PVNL', 'PVNR',
        # 'RID', 'RIML', 'RIMR', 'RIVL', 'RIVR',
        # 'RMDDL', 'RMDDR', 'RMDL', 'RMDR', 'RMDVL', 'RMDVR', 'RMED', 'RMEL', 'RMER', 'RMEV', 'RMFL', 'RMGL', 'RMGR',
        # 'RMHL', 'RMHR',
        # 'SMBDL', 'SMBDR', 'SMBVL', 'SMBVR', 'SMDDL', 'SMDDR', 'SMDVL', 'SMDVR',
        # 'URADL', 'URADR', 'URAVL', 'URAVR',
        # 'VC1', 'VC2', 'VC3', 'VC4', 'VC5', 'VC6',
    ]

    # cells = None
    # cells_to_stimulate = ['PLML', 'PLMR', 'AVM']
    cells_to_stimulate = []

    # cells_to_plot = ['AVAL', 'AVAR', 'AVBL', 'AVBR', 'PLML', 'PLMR', 'AVM', 'ALML', 'ALMR', 'PVDL', 'PVDR', 'AVDL', 'AVDR']
    # cells_to_plot = ['PLML', 'AVM', 'AVBL', 'AVAL']
    # cells_to_plot = cells
    # cells_to_plot += motors
    reference = "c302_%s_TapWithdrawal" % parameter_set

    conn_polarity_override = {
        "ALML-ALML": "inh",
        "ALML-PVCL": "inh",
        "ALML-PVCR": "inh",
        "ALML-AVDR": "inh",  ##
        "ALMR-PVCR": "inh",
        "AVM-PVCL": "inh",
        "AVM-PVCR": "inh",
        "AVM-AVBL": "inh",
        "AVM-AVBR": "inh",
        "AVM-AVDL": "inh",  ##
        "AVM-AVDR": "inh",  ##
        "PVDL-PVDR": "inh",
        "PVDL-PVCL": "exc",
        "PVDL-PVCR": "exc",
        "PVDL-AVAL": "inh",
        "PVDL-AVAR": "inh",
        "PVDL-AVDL": "inh",
        "PVDL-AVDR": "inh",
        "PVDR-PVDL": "inh",
        "PVDR-DVA": "exc",
        "PVDR-PVCL": "exc",
        "PVDR-PVCR": "exc",
        "PVDR-AVAL": "inh",
        "PVDR-AVAR": "inh",
        "PVDR-AVDL": "inh",
        "DVA-PVCL": "inh",  #
        "DVA-PVCR": "inh",  #
        "DVA-AVAL": "inh",
        "DVA-AVAR": "inh",
        "DVA-AVBL": "inh",  #
        "DVA-AVBR": "inh",  #
        "DVA-AVDR": "inh",
        "PVCL-DVA": "exc",
        "PVCL-PVCL": "inh",
        "PVCL-PVCR": "inh",
        "PVCL-AVAL": "inh",
        "PVCL-AVAR": "inh",
        "PVCL-AVBL": "exc",
        "PVCL-AVBR": "exc",
        "PVCL-AVDL": "inh",
        "PVCL-AVDR": "inh",
        "PVCR-PVDL": "inh",
        "PVCR-PVDR": "inh",
        "PVCR-DVA": "exc",
        "PVCR-PVCL": "inh",
        "PVCR-AVAL": "inh",
        "PVCR-AVAR": "inh",
        "PVCR-AVBL": "exc",
        "PVCR-AVBR": "exc",
        "PVCR-AVDL": "inh",
        "PVCR-AVDR": "inh",
        "AVAL-PVCL": "inh",
        "AVAL-PVCR": "inh",
        "AVAL-AVAR": "inh",
        "AVAL-AVBL": "inh",
        "AVAL-AVDL": "inh",
        "AVAL-AVDR": "inh",
        "AVAR-PVCL": "inh",
        "AVAR-PVCR": "inh",
        "AVAR-AVAL": "inh",
        "AVAR-AVBL": "inh",
        "AVAR-AVBR": "inh",
        "AVAR-AVDL": "inh",
        "AVAR-AVDR": "inh",
        "AVBL-DVA": "inh",
        "AVBL-PVCR": "inh",
        "AVBL-AVAL": "inh",
        "AVBL-AVAR": "inh",
        "AVBL-AVBR": "inh",
        "AVBL-AVDR": "inh",
        "AVBR-AVAL": "inh",
        "AVBR-AVAR": "inh",
        "AVBR-AVBL": "inh",
        "AVBR-AVDL": "inh",
        "AVDL-PVCL": "inh",
        "AVDL-AVAL": "exc",
        "AVDL-AVAR": "exc",
        "AVDL-AVDR": "exc",
        "AVDR-PVCR": "inh",
        "AVDR-AVAL": "exc",
        "AVDR-AVAR": "exc",
        "AVDR-AVBL": "inh",
        "AVDR-AVDL": "exc",
        "DA9-DVA": "inh",
        "DVA-DA2": "inh",
        "PVCL-DA5": "inh",
        "PVCL-DA8": "inh",
        "PVCR-DA2": "inh",
        "PVCR-DA4": "inh",
        "PVCR-DA5": "inh",
        "PVCR-DA7": "inh",
        "AVBL-DA5": "inh",
        "AVBL-DA7": "inh",
        "AVBR-DA3": "inh",
        "AVBR-DA4": "inh",
        "AVBR-DA7": "inh",
        "AVDL-DA1": "inh",
        "AVDL-DA2": "inh",
        "AVDL-DA3": "inh",
        "AVDL-DA4": "inh",
        "AVDL-DA5": "inh",
        "AVDL-DA8": "inh",
        "AVDR-DA1": "inh",
        "AVDR-DA2": "inh",
        "AVDR-DA3": "inh",
        "AVDR-DA4": "inh",
        "AVDR-DA5": "inh",
        "AVDR-DA8": "inh",
        "DB1-DA1": "inh",
        "DB1-DA2": "inh",
        "DB2-DA2": "inh",
        "DB2-DA3": "inh",
        "DB2-DA4": "inh",
        "DB3-DA4": "inh",
        "DB3-DA5": "inh",
        "DB4-DA5": "inh",
        "DB5-DA6": "inh",
        "DB5-DA7": "inh",
        "DB5-DA8": "inh",
        "DB6-DA8": "inh",
        "DB6-DA9": "inh",
        "DB7-DA9": "inh",
        "AVAR-DB2": "inh",
        "AVAR-DB3": "inh",
        "AVAL-DB7": "inh",
        "DA1-DB1": "inh",
        "DA2-DA3": "inh",
        "DA2-DB1": "inh",
        "DA3-DB3": "inh",
        "DA4-DB2": "inh",
        "DA5-DB4": "inh",
        "DA6-DB5": "inh",
        "DA7-DB6": "inh",
        "DA8-DB7": "inh",
        "DA9-DB7": "inh",
        "DVA-VA2": "inh",
        "DVA-VA6": "inh",
        "DVA-VA8": "inh",
        "DVA-VA12": "inh",
        "PVCL-VA11": "inh",
        "PVDR-VA9": "inh",
        "PVDR-VA12": "inh",
        "AVBL-VA2": "inh",
        "AVBL-VA10": "inh",
        "AVBR-VA3": "inh",
        "AVBR-VA4": "inh",
        "AVDL-VA3": "inh",
        "AVDL-VA5": "inh",
        "AVDL-VA10": "inh",
        "AVDL-VA12": "inh",
        "AVDR-VA2": "inh",
        "AVDR-VA3": "inh",
        "AVDR-VA5": "inh",
        "AVDR-VA11": "inh",
        "VB1-VA1": "inh",
        "VB1-VA2": "inh",
        "VB1-VA3": "inh",
        "VB1-VA4": "inh",
        "VB2-VA2": "inh",
        "VB2-VA3": "inh",
        "VB3-VA4": "inh",
        "VB3-VA5": "inh",
        "VB4-VA4": "inh",
        "VB4-VA5": "inh",
        "VB5-VA6": "inh",
        "VB6-VA7": "inh",
        "VB6-VA8": "inh",
        "VB7-VA9": "inh",
        "VB7-VA10": "inh",
        "VB8-VA11": "inh",
        "VB9-VA11": "inh",
        "VB10-VA11": "inh",
        "VB11-VA12": "inh",
        "VB11-PVCR": "inh",
        "VB4-VB5": "inh",
        "VA2-VB1": "inh",
        "VA2-VB2": "inh",
        "VA3-VB2": "inh",
        "VA3-VB3": "inh",
        "VA4-AVDL": "inh",
        "VA4-VB3": "inh",
        "VA4-VB4": "inh",
        "VA5-VB4": "inh",
        "VA6-VB4": "inh",
        "VA6-VB5": "inh",
        "VA7-VB6": "inh",
        "VA8-VB6": "inh",
        "VA9-VB7": "inh",
        "VA9-VB8": "inh",
        "VA10-VB8": "inh",
        "VA10-VB9": "inh",
        "VA11-VB10": "inh",
        "VA12-PVCL": "inh",
        "VA12-PVCR": "inh",
        "VA12-DB7": "inh",
        "VA12-VB11": "inh",
        "AVM-VB3": "inh",
        #'VB4-VB5':'exc',
        #'DD1-DA2':'inh',
        #'DD1-VB2':'inh',
        #'DD2-DA3':'inh',
        #'DD3-DA5':'inh',
    }

    conn_number_override = {
        #'PVCL-AVDL':7*0.1,
        #'PVCL-AVDR':11*0.1,
        #'PVCR-AVDL':16*0.1,
        #'PVCR-AVDR':6*0.1,
        #'PVCR-AVDR_GJ':2 * 0.01,
        #'AVDR-PVCR_GJ':2 * 0.01,
        "PVCL-AVAL_GJ": 5 * 0.01,
        "AVAL-PVCL_GJ": 5 * 0.01,
        "PVCL-AVAR_GJ": 10 * 0.01,
        "AVAR-PVCL_GJ": 10 * 0.01,
        "PVCR-AVAL_GJ": 15 * 0.01,
        "AVAL-PVCR_GJ": 15 * 0.01,
        "PVCR-AVAR_GJ": 22 * 0.01,
        "AVAR-PVCR_GJ": 22 * 0.01,
        "PVCL-PLML_GJ": 4 * 0.01,  ##
        "PVCR-PLMR_GJ": 8 * 0.01,  ##
        #'AVDL-AVM_GJ':8 * 0.01,
        #'ALML-AVM_GJ':1 * 0.01,
        #'ALMR-AVM_GJ':1 * 0.01,
        #'AVDR-ALMR_GJ':1 * 0.01,
        # 'AVDL-AVAL':37*0.1,
        # 'AVDL-AVAR':37*0.1,
        # 'AVDR-AVAL':41*0.1,
        # 'AVDR-AVAR':52*0.1,
        # 'AVDL-AVAL_GJ':7*0.1,
        # 'AVAL-AVDL_GJ':7*0.1,
        # 'AVDL-AVAR_GJ':2*0.1,
        # 'AVAR-AVDL_GJ':2*0.1,
        # 'AVDR-AVAL_GJ':9*0.1,
        # 'AVAL-AVDR_GJ':9*0.1,
        # 'AVDR-AVAR_GJ':15*0.1,
        # 'AVAR-AVDR_GJ':15*0.1,
        # 'ALMR-AVDR_GJ':2*5,
        # 'AVDR-ALMR_GJ':2*5,
        "AVAR-AVBL_GJ": 3 * 0.01,
        "AVBL-AVAR_GJ": 3 * 0.01,
        # 'AVAR-AVAL_GJ':18*2,
        # 'AVAL-AVAR_GJ':18*2,
        "PVDL-AVAR_GJ": 4 * 0.01,
        "AVAR-PVDL_GJ": 4 * 0.01,
        "PVDR-AVAL_GJ": 6 * 0.01,
        "AVAL-PVDR_GJ": 6 * 0.01,
        "AVBL-VA11_GJ": 1 * 0.01,
        "VA11-AVBL_GJ": 1 * 0.01,
        "AVBR-VA11_GJ": 3 * 0.01,
        "VA11-AVBR_GJ": 3 * 0.01,
        "PVCR-VA11_GJ": 3 * 0.01,
        "VA11-PVCR_GJ": 3 * 0.01,
        "DVA-VA11_GJ": 1 * 0.01,
        "VA11-DVA_GJ": 1 * 0.01,
        "PVCL-VA12_GJ": 18 * 0.01,
        "VA12-PVCL_GJ": 18 * 0.01,
        "PVCR-VA12_GJ": 8 * 0.01,
        "VA12-PVCR_GJ": 8 * 0.01,
        "AVAL-VB11_GJ": 2 * 0.01,
        "VB11-AVAL_GJ": 2 * 0.01,
        "PVCL-DA4_GJ": 1 * 0.01,
        "DA4-PVCL_GJ": 1 * 0.01,
        "PVCL-DA7_GJ": 1 * 0.01,
        "DA7-PVCL_GJ": 1 * 0.01,
        "PVCR-DA7_GJ": 3 * 0.01,
        "DA7-PVCR_GJ": 3 * 0.01,
        "PVCL-DA8_GJ": 17 * 0.01,
        "DA8-PVCL_GJ": 17 * 0.01,
        "PVCR-DA8_GJ": 1 * 0.01,
        "DA8-PVCR_GJ": 1 * 0.01,
        "DVA-DA9_GJ": 3 * 0.01,
        "DA9-DVA_GJ": 3 * 0.01,
        "PVCR-DA9_GJ": 3 * 0.01,
        "DA9-PVCR_GJ": 3 * 0.01,
        "DB7-VA10_GJ": 1 * 0.01,
        "VA10-DB7_GJ": 1 * 0.01,
        "VA4-VB3_GJ": 1 * 0.01,
        "VB3-VA4_GJ": 1 * 0.01,
        "VA11-VB10_GJ": 3 * 0.01,
        "VB10-VA11_GJ": 3 * 0.01,
        "VA12-VB11_GJ": 7 * 0.01,
        "VB11-VA12_GJ": 7 * 0.01,
        "VB11-DA9_GJ": 7 * 0.01,
        "DA9-VB11_GJ": 7 * 0.01,
    }

    nml_doc = None

    if generate:
        nml_doc = c302.generate(
            reference,
            params,
            cells=cells,
            cells_to_plot=cells_to_plot,
            cells_to_stimulate=cells_to_stimulate,
            conn_polarity_override=conn_polarity_override,
            conn_number_override=conn_number_override,
            muscles_to_include=muscles_to_include,
            duration=duration,
            dt=dt,
            target_directory=target_directory,
            data_reader=data_reader,
            param_overrides=param_overrides,
            verbose=verbose,
        )

        # stim_amplitude = "6pA"
        # stim_amplitude = "5.135697186048022pA"

        for vb in VB_motors:
            c302.add_new_sinusoidal_input(
                nml_doc,
                cell=vb,
                delay="0ms",
                duration="1000ms",
                amplitude="3pA",
                period="150ms",
                params=params,
            )

        for db in DB_motors:
            c302.add_new_sinusoidal_input(
                nml_doc,
                cell=db,
                delay="0ms",
                duration="1000ms",
                amplitude="3pA",
                period="150ms",
                params=params,
            )

        # c302.add_new_input(nml_doc, "AVM", "10ms", "700ms", stim_amplitude, params)
        # c302.add_new_input(nml_doc, "ALML", "10ms", "700ms", stim_amplitude, params)
        # c302.add_new_input(nml_doc, "ALMR", "10ms", "700ms", stim_amplitude, params)
        # c302.add_new_input(nml_doc, "PLML", "10ms", "700ms", stim_amplitude, params)
        # c302.add_new_input(nml_doc, "PLMR", "10ms", "700ms", stim_amplitude, params)

        nml_file = target_directory + "/" + reference + ".net.nml"
        writers.NeuroMLWriter.write(
            nml_doc, nml_file
        )  # Write over network file written above...

        c302.print_("(Re)written network file to: " + nml_file)

    return cells, cells_to_stimulate, params, muscles_to_include, nml_doc


if __name__ == "__main__":
    parameter_set = sys.argv[1] if len(sys.argv) == 2 else "C2"

    setup(parameter_set, generate=True, data_reader=c302.FW_DATA_READER)



#! /usr/bin/env python

from neuroml import NeuroMLDocument
from neuroml import Network
from neuroml import Population
from neuroml import Instance
from neuroml import IncludeType
from neuroml import Location
from neuroml import Input
from neuroml import InputList
from neuroml import Projection
from neuroml import Connection
from neuroml import ConnectionWD
from neuroml import ElectricalProjection
from neuroml import ElectricalConnectionInstanceW
from neuroml import ContinuousProjection
from neuroml import ContinuousConnectionInstanceW
from neuroml import Property
from neuroml import PulseGenerator
from neuroml import SineGenerator
from neuroml import SilentSynapse

import neuroml.writers as writers
import neuroml.loaders as loaders

import cect

# import c302.bioparameters

import airspeed

import random
import argparse
import shutil
import os
import logging
import importlib
import math
from lxml import etree
import re
import json
import collections

try:
    from owmeta_core import __version__ as owc_version
    from owmeta_core.bundle import Bundle
    from owmeta_core.context import Context
    from owmeta import __version__ as owmeta_version
    from owmeta.cell import Cell
    from owmeta.neuron import Neuron
    from owmeta.muscle import Muscle

    owmeta_installed = True

except Exception:
    print("owmeta not installed! Proceeding anyway...")
    owmeta_installed = False

try:
    from urllib2 import URLError  # Python 2
except Exception:
    from urllib.error import URLError  # Python 3


from cect.Cells import BODY_WALL_MUSCLE_NAMES

DEFAULT_DATA_READER = "SpreadsheetDataReader"
#
# DEFAULT_DATA_READER = "cect.White_whole"
# DEFAULT_DATA_READER = "cect.Cook2019HermReader"
DEFAULT_DATA_READER = "cect.SpreadsheetDataReader"
FW_DATA_READER = "cect.UpdatedSpreadsheetDataReader2"

logging.basicConfig()

here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(here, "__version__.py")) as version_file:
    exec(version_file.read(), about)
__version__ = about["__version__"]

LEMS_TEMPLATE_FILE = "LEMS_c302_TEMPLATE.xml"

MUSCLE_RE = re.compile(r"M([VD][LR])(\d+)")

OWMETA_CACHED_DATA_FILE = (
    os.path.dirname(os.path.abspath(__file__)) + "/data/owmeta_cache.json"
)


def print_(msg, print_it=True):  # print_it=False when not verbose
    if print_it:
        pre = "c302      >>> "
        print("%s %s" % (pre, msg.replace("\n", "\n" + pre)))


def load_data_reader(data_reader):
    """
    Imports and returns data reader module
    Args:
        data_reader (str): The name of the data reader
    Returns:
        reader (obj): The data reader object
    """
    # return importlib.import_module("c302.%s" % data_reader)
    if "cect" in data_reader:
        dr = importlib.import_module("%s" % data_reader)
        return dr.get_instance()
    else:
        return importlib.import_module("c302.%s" % data_reader)


def get_str_from_expnotation(num):
    """
    Returns a formatted string representing a floating point number, e.g. 1*0.00001 would result into 1e-05. Returning 0.00001.
    Args:
        num (float): A number. Can be of type int or float,  float can have exponential notation.
    Returns:
       (str): A string representing a float with 15 fractional digits.
    """
    return "{0:.15f}".format(num)


def get_muscle_position(muscle, data_reader):
    if muscle == "MANAL" or muscle == "MVULVA":
        return 0, 0, 0
    # TODO: Pull these positions from openworm/owmeta-data
    pat1 = r"M([VD])([LR])(\d+)"
    pat2 = r"([VD])([LR])(\d+)"
    md = re.fullmatch(pat1, muscle)
    if not md:
        md = re.fullmatch(pat2, muscle)

    if md:
        dv = md.group(1)
        lr = md.group(2)
        idx = md.group(3)
        x = 80 * (1 if lr == "L" else -1)
        z = 80 * (-1 if dv == "V" else 1)
        y = -300 + 30 * int(idx)
        return x, y, z

    raise Exception("Unrecognized muscle name format %s" % muscle)


def is_muscle(cell_name):
    return MUSCLE_RE.fullmatch(cell_name)


def process_args():
    """
    Parse command-line arguments.
    """
    parser = argparse.ArgumentParser(
        "c302",
        description="A script which can generate NeuroML2 compliant networks based on the C elegans connectome, along with LEMS files to run them",
    )

    parser.add_argument(
        "reference",
        type=str,
        metavar="<reference>",
        help="Unique reference for new network",
    )

    parser.add_argument(
        "parameters",
        type=str,
        metavar="<parameters>",
        help="Set of biophysical parametes to use, e.g. parameters_A",
    )

    parser.add_argument(
        "-datareader",
        type=str,
        metavar="<data-reader>",
        default=DEFAULT_DATA_READER,
        help="Use a specific data reader. (default: %s)" % DEFAULT_DATA_READER,
    )

    parser.add_argument(
        "-cells",
        type=str,
        metavar="<cells>",
        default=None,
        help="List of cells to include in network (default: all)",
    )

    parser.add_argument(
        "-cellstoplot",
        type=str,
        metavar="<cells-to-plot>",
        default=None,
        help="List of cells to plot (default: all)",
    )

    parser.add_argument(
        "-cellstostimulate",
        type=str,
        metavar="<cells-to-stimulate>",
        default=None,
        help="List of cells to stimulate (default: all)",
    )

    parser.add_argument(
        "-connpolarityoverride",
        type=str,
        metavar="<conn-polarity-override>",
        default=None,
        help='Map of connection polarities to override, e.g. {"AVAL-AVBR":"inh", ...} => use inhibitory connection for AVAL-AVBR',
    )

    parser.add_argument(
        "-connnumberoverride",
        type=str,
        metavar="<conn-number-override>",
        default=None,
        help='Map of connection numbers to override, e.g. {"I1L-I3":2.5, "AVAR-AVBL_GJ":2} => use 2.5 connections from I1L to I3, use 2 connections for GJ AVAR-AVBL',
    )

    parser.add_argument(
        "-paramoverride",
        type=str,
        metavar="<param-override>",
        default=None,
        help='Map of parameters to override, e.g. {"unphysiological_offset_current":"2pA", ...} => use 2pA for additional offset currnet',
    )

    parser.add_argument(
        "-connnumberscaling",
        type=str,
        metavar="<conn-number-scaling>",
        default=None,
        help='Map of scaling factors for connection numbers, e.g. {"I1L-I3":2, "AVAR-AVBL_GJ":2} => use 2 times as many connections from I1L to I3, use 2 times as many connections for GJ AVAR-AVBL',
    )

    parser.add_argument(
        "-musclestoinclude",
        type=str,
        metavar="<muscles-to-include>",
        default=[],
        help="List of muscles to include (default: empty list, i.e. none)",
    )

    parser.add_argument(
        "-duration",
        type=float,
        metavar="<duration>",
        default=100,
        help="Duration of simulation in ms",
    )

    parser.add_argument(
        "-dt",
        type=float,
        metavar="<time step>",
        default=0.01,
        help="Timestep for simulations (dt) in ms",
    )

    parser.add_argument(
        "-vmin",
        type=float,
        metavar="<vmin>",
        default=-80,
        help="Minimum voltage for plot in mV",
    )

    parser.add_argument(
        "-vmax",
        type=float,
        metavar="<vmax>",
        default=-40,
        help="Maximum voltage for plot in mV",
    )

    return parser.parse_args()


quadrant0 = "MDR"
quadrant1 = "MVR"
quadrant2 = "MVL"
quadrant3 = "MDL"

# soma positions from http://www.wormatlas.org/neuronalwiring.html - 2.2 Neuron Description (Neuron Types)
VB_soma_pos = {
    "VB1": 0.21,
    "VB2": 0.19,
    "VB3": 0.28,
    "VB4": 0.32,
    "VB5": 0.38,
    "VB6": 0.45,
    "VB7": 0.5,
    "VB8": 0.57,
    "VB9": 0.61,
    "VB10": 0.67,
    "VB11": 0.72,
}

DB_soma_pos = {
    "DB1": 0.24,
    "DB2": 0.21,
    "DB3": 0.3,
    "DB4": 0.39,
    "DB5": 0.51,
    "DB6": 0.62,
    "DB7": 0.72,
}


def get_next_stim_id(nml_doc, cell):
    i = 1
    for stim in nml_doc.pulse_generators:
        if stim.id.startswith("%s_%s" % ("stim", cell)):
            i += 1
    id = "%s_%s_%s" % ("stim", cell, i)
    return id


def get_cell_position(cell):
    root_dir = os.path.dirname(os.path.abspath(__file__))
    # cell_file_path = root_dir + "/../../../" if test else root_dir + "/../../"  # if running test
    cell_file_path = root_dir + "/"
    cell_file = cell_file_path + "NeuroML2/%s.cell.nml" % cell
    doc = loaders.NeuroMLLoader.load(cell_file)
    location = doc.cells[0].morphology.segments[0].proximal
    # print "%s, %s, %s" %(location.x, location.y, location.z)
    return location


def append_input_to_nml_input_list(stim, nml_doc, cell, params):
    target = get_cell_id_string(cell, params, muscle=is_muscle(cell))

    input_list = InputList(
        id="Input_%s_%s" % (cell, stim.id), component=stim.id, populations="%s" % cell
    )

    input_list.input.append(Input(id=0, target=target, destination="synapses"))

    nml_doc.networks[0].input_lists.append(input_list)


def add_new_sinusoidal_input(nml_doc, cell, delay, duration, amplitude, period, params):
    id = get_next_stim_id(nml_doc, cell)

    if cell.startswith("VB"):
        phase = VB_soma_pos[cell]
    else:
        phase = DB_soma_pos[cell]
    # phase = get_cell_position(cell).x
    phase = phase * -0.886
    print_("### CELL %s PHASE: %s" % (cell, phase))

    if cell.startswith("VB"):
        if amplitude.startswith("-"):
            amplitude = amplitude[1:]
        else:
            amplitude = "-" + amplitude

    input = SineGenerator(
        id=id,
        delay=delay,
        phase=phase,
        duration=duration,
        amplitude=amplitude,
        period=period,
    )
    nml_doc.sine_generators.append(input)

    append_input_to_nml_input_list(input, nml_doc, cell, params)


def add_new_input(nml_doc, cell, delay, duration, amplitude, params):
    id = get_next_stim_id(nml_doc, cell)
    input = PulseGenerator(id=id, delay=delay, duration=duration, amplitude=amplitude)
    nml_doc.pulse_generators.append(input)

    append_input_to_nml_input_list(input, nml_doc, cell, params)


def get_muscle_names():
    names = []
    for i in range(24):
        names.append("%s%s" % (quadrant0, i + 1 if i > 8 else ("0%i" % (i + 1))))
    for i in range(24):
        names.append("%s%s" % (quadrant1, i + 1 if i > 8 else ("0%i" % (i + 1))))
    for i in range(24):
        names.append("%s%s" % (quadrant2, i + 1 if i > 8 else ("0%i" % (i + 1))))
    for i in range(24):
        names.append("%s%s" % (quadrant3, i + 1 if i > 8 else ("0%i" % (i + 1))))

    return names


def merge_with_template(model, templfile):
    with open(templfile) as f:
        templ = airspeed.Template(f.read())
    return templ.merge(model)


def write_to_file(
    nml_doc,
    lems_info,
    reference,
    template_path="",
    validate=True,
    verbose=True,
    target_directory=".",
):
    #######   Write to file  ######

    nml_file = target_directory + "/" + reference + ".net.nml"
    print_("Writing generated network to: %s" % os.path.realpath(nml_file))
    writers.NeuroMLWriter.write(nml_doc, nml_file)

    if verbose:
        print_("Written network file to: " + nml_file)

    lems_file_name = target_directory + "/" + "LEMS_%s.xml" % reference
    with open(lems_file_name, "w") as lems:
        # if running unittest concat template_path
        merged = merge_with_template(lems_info, template_path + LEMS_TEMPLATE_FILE)
        lems.write(merged)

    if verbose:
        print_("Written LEMS file to: " + lems_file_name)

    if validate:
        ###### Validate the NeuroML ######

        from neuroml.utils import validate_neuroml2

        try:
            validate_neuroml2(nml_file)
        except URLError:
            print_("Problem validating against remote Schema!")


# Get the standard name for a network connection
def get_projection_id(pre, post, synclass, syntype):
    proj_id = "NC_%s_%s_%s" % (pre, post, synclass)
    """
    if "GapJunction" in syntype:
       proj_id += '_GJ' """

    return proj_id


def get_random_colour_hex():
    rgb = [
        hex(random.randint(0, 256)),
        hex(random.randint(0, 256)),
        hex(random.randint(0, 256)),
    ]
    col = "#"
    for c in rgb:
        col += c[2:4] if len(c) == 4 else "00"
    return col


def get_file_name_relative_to_c302(file_name):
    if "C302_HOME" in os.environ:
        return os.path.relpath(os.environ["C302_HOME"], file_name)


def get_cell_names_and_connection(data_reader, test=False):
    # Use the data reader to give a list of all cells and a list of all connections
    # This could be replaced with a call to "DatabaseReader" or "OpenWormNeuroLexReader" in future...
    # If called from unittest folder ammend path to "../../../../"

    cell_names, conns = load_data_reader(data_reader).read_data(
        include_nonconnected_cells=True
    )

    cell_names.sort()

    return cell_names, conns


def get_cell_muscle_names_and_connection(data_reader, test=False):
    mneurons, all_muscles, muscle_conns = load_data_reader(
        data_reader
    ).read_muscle_data()
    if "MANAL" in all_muscles:
        all_muscles.remove("MANAL")
    if "MVULVA" in all_muscles:
        all_muscles.remove("MVULVA")

    all_known_muscles = []
    if len(all_muscles) == 0:
        all_known_muscles = BODY_WALL_MUSCLE_NAMES
    else:
        for m in all_muscles:
            if m in BODY_WALL_MUSCLE_NAMES:
                all_known_muscles.append(m)

    all_known_muscles = sorted(all_known_muscles)

    print_(" - Using for muscles: %s" % all_known_muscles)

    return mneurons, all_known_muscles, muscle_conns


def is_cond_based_cell(params):
    return params.is_level_C() or params.is_level_D()


def get_cell_id_string(cell, params, muscle=False):
    if cell in get_muscle_names():
        muscle = True
    if not params.is_level_D():
        if not muscle:
            return "../%s/0/%s" % (cell, params.generic_neuron_cell.id)
        else:
            return "../%s/0/%s" % (cell, params.generic_muscle_cell.id)

    else:
        if not muscle:
            return "../%s/0/%s" % (cell, cell)
        else:
            return "../%s/0/%s" % (cell, params.generic_muscle_cell.id)


def regex_match(pattern, str):
    return is_regex_string(pattern) and re.match(pattern, str)


def is_regex_string(str):
    return "^" in str and "$" in str


def elem_in_coll_matches_conn(coll, conn):
    for elem in coll:
        if regex_match(elem, conn):
            return True
    return False


cached_owmeta_data = None


def _get_cell_info(bnd, cells):
    global cached_owmeta_data
    # print('------ Getting the cell info for %s'%cells)
    all_neuron_info = collections.OrderedDict()
    all_muscle_info = collections.OrderedDict()

    if bnd is None:
        if cached_owmeta_data is None:
            print_("Loading owmeta cached data from: %s" % OWMETA_CACHED_DATA_FILE)
            with open(OWMETA_CACHED_DATA_FILE) as f:
                cached_owmeta_data = json.load(f)

        for cell in cells:
            if is_muscle(cell):
                all_muscle_info[cell] = cached_owmeta_data["muscle_info"][cell]
            else:
                all_neuron_info[cell] = cached_owmeta_data["neuron_info"][cell]

    else:
        ctx = bnd(Context)(ident="http://openworm.org/data").stored
        # Go through our list and get the neuron object associated with each name.
        # Store these in another list.
        fixed_up_names = []
        for name in cells:
            match = is_muscle(name)
            if match:
                name = match.group(1) + str(int(match.group(2)))
            fixed_up_names.append(name)
        # fixed_up_names.remove('MANAL')
        # fixed_up_names.remove('MVULVA')

        for name in fixed_up_names:
            cell = next(ctx(Cell).query(name=name).load(), None)
            if cell is None:
                # print_("No matching cell for %s" % name)
                continue

            normalized_name = cell.name()
            short = ") %s" % normalized_name
            color = ".5 0 0"
            if isinstance(cell, Neuron):
                neuron_types = cell.type()
                if "sensory" in neuron_types:
                    short = "Se%s" % short
                    color = "1 .2 1"
                if "interneuron" in neuron_types:
                    short = "In%s" % short
                    color = "1 0 .4"
                if "motor" in neuron_types:
                    short = "Mo%s" % short
                    color = ".5 .4 1"

                neurotransmitter = cell.neurotransmitter()
            elif isinstance(cell, Muscle):
                neuron_types = ()
                neurotransmitter = ()
                color = "0 0.6 0"
                short = "Mu%s" % short
            else:
                # At this point, we should only have Neurons and Muscles because the reader
                # filters them out
                raise Exception("Got an unexpected cell type")

            short = "(%s" % short
            receptor = cell.receptor()
            if "GABA" in neurotransmitter:
                short = "- %s" % short
            elif len(neurotransmitter) == 0:
                short = "? %s" % short
            else:
                short = "+ %s" % short

            info = (cell, neuron_types, receptor, neurotransmitter, short, color)

            if isinstance(cell, Muscle):
                all_muscle_info[cell.name()] = info
            elif isinstance(cell, Neuron):
                all_neuron_info[cell.name()] = info

    # print('==== Returning %s; %s'%(all_neuron_info, all_muscle_info))
    return all_neuron_info, all_muscle_info


def set_param(params, param, value):
    v = params.get_bioparameter(param, warn_if_missing=False)
    if v:
        if v == value:
            return
        print_("Setting parameter %s = %s" % (param, value))
        params.set_bioparameter(param, value, "Set with param_overrides", 0)
    else:
        print_("Adding parameter %s = %s" % (param, value))
        params.add_bioparameter(param, value, "Add with param_overrides", 0)


def mirror_param(params, k, v):
    pattern = k.split("_")
    pre = pattern[0]
    pattern[0] = "%s"
    post = pattern[2]
    pattern[2] = "%s"
    tmp_param = pattern[5]
    pattern[5] = "%s"
    pattern = "_".join(pattern)

    override_key1 = pattern % (pre, post, tmp_param)
    override_key2 = pattern % (post, pre, tmp_param)

    set_param(params, override_key1, v)
    set_param(params, override_key2, v)


def generate(
    net_id,
    params,
    data_reader=DEFAULT_DATA_READER,
    cells=None,
    cells_to_plot=None,
    cells_to_stimulate=None,
    muscles_to_include=[],
    conns_to_include=[],
    conns_to_exclude=[],
    conn_number_override=None,
    conn_number_scaling=None,
    conn_polarity_override=None,
    duration=500,
    dt=0.01,
    vmin=None,
    vmax=None,
    seed=1234,
    test=False,
    verbose=True,
    print_connections=False,
    param_overrides={},
    target_directory="./",
):
    validate = not (
        params.is_level_B()
        or params.is_level_C0()
        or params.is_level_C2
        or params.is_level_D1()
    )

    root_dir = os.path.dirname(os.path.abspath(__file__))

    regex_param_overrides = {"mirrored_elec_conn_params": {}}
    if param_overrides:
        for k, v in param_overrides.items():
            if k == "mirrored_elec_conn_params":
                for mk, mv in v.items():
                    if is_regex_string(mk):
                        regex_param_overrides["mirrored_elec_conn_params"][mk] = mv
                        continue
                    else:
                        mirror_param(params, mk, mv)

            elif k == "custom_component_type_gate_overrides":
                continue  # channel params

            elif is_regex_string(k):
                regex_param_overrides[k] = v
                continue  # Add specific param later (e.g. add 'AVB.-DB\d+_elec_syn_gbase' during connection parsing)
            else:
                set_param(params, k, v)

    params.create_models()

    if vmin is None:
        if params.is_level_A() or params.is_level_B():
            vmin = -52
        elif params.is_level_C():
            vmin = -60
        elif params.is_level_D():
            vmin = -60
        else:
            vmin = -52

    if vmax is None:
        if params.is_level_A() or params.is_level_B():
            vmax = -28
        elif params.is_level_C():
            vmax = 25
        elif params.is_level_D():
            vmax = 25
        else:
            vmax = -28

    random.seed(seed)

    info = (
        "\n\nParameters and setting used to generate this network:\n\n"
        + "    Data reader:                    %s\n" % data_reader
        + "    c302 version:                   %s\n" % __version__
        + (
            "    cect version:                   %s\n" % (cect.__version__)
            if "cect" in data_reader
            else ""
        )
        + "    owmeta version:                 %s\n"
        % ("- not installed -" if not owmeta_installed else owmeta_version)
        + "    owmeta_core version:            %s\n"
        % ("- not installed -" if not owmeta_installed else owc_version)
        + "    Cells:                          %s\n"
        % (cells if cells is not None else "All cells")
        + "    Cell stimulated:                %s\n"
        % (cells_to_stimulate if cells_to_stimulate is not None else "All neurons")
        + "    Connection:                     %s\n"
        % (conns_to_include if conns_to_include is not None else "All connections")
        + "    Connection numbers overridden:  %s\n"
        % (conn_number_override if conn_number_override is not None else "None")
        + "    Connection numbers scaled:      %s\n"
        % (conn_number_scaling if conn_number_scaling is not None else "None")
        + "    Connection polarities override: %s\n" % conn_polarity_override
        + "    Muscles:                        %s\n"
        % (muscles_to_include if muscles_to_include is not None else "All muscles")
    )

    info_settings = info
    if verbose:
        print_(info)
    info += "\n%s\n" % (params.bioparameter_info("    "))

    nml_doc = NeuroMLDocument(id=net_id, notes=info)

    if params.is_level_A() or params.is_level_B() or params.level == "BC1":
        nml_doc.iaf_cells.append(params.generic_muscle_cell)
        nml_doc.iaf_cells.append(params.generic_neuron_cell)
    elif params.is_level_C():
        nml_doc.cells.append(params.generic_muscle_cell)
        nml_doc.cells.append(params.generic_neuron_cell)
    elif params.is_level_D():
        nml_doc.cells.append(params.generic_muscle_cell)
    elif params.is_level_X():
        nml_doc.cells.append(params.generic_muscle_cell)
        nml_doc.cells.append(params.generic_neuron_cell)

    net = Network(id=net_id)

    nml_doc.networks.append(net)

    net.properties.append(Property("recommended_duration_ms", duration))
    net.properties.append(Property("recommended_dt_ms", dt))

    nml_doc.pulse_generators.append(params.offset_current)

    if is_cond_based_cell(params):
        if isinstance(params.concentration_model, list):
            nml_doc.fixed_factor_concentration_models.extend(params.concentration_model)
        else:
            nml_doc.fixed_factor_concentration_models.append(params.concentration_model)

    cell_names, conns = get_cell_names_and_connection(data_reader)

    # To hold all Cell NeuroML objects vs. names
    all_cells = {}

    # lems_file = ""
    lems_info = {
        "comment": info,
        "reference": net_id,
        "duration": duration,
        "dt": dt,
        "vmin": vmin,
        "vmax": vmax,
    }

    lems_info["plots"] = []
    lems_info["activity_plots"] = []
    lems_info["muscle_plots"] = []
    lems_info["muscle_activity_plots"] = []

    lems_info["to_save"] = []
    lems_info["activity_to_save"] = []
    lems_info["muscles_to_save"] = []
    lems_info["muscles_activity_to_save"] = []
    lems_info["cells"] = []
    lems_info["muscles"] = []
    lems_info["includes"] = []

    if params.custom_component_types_definitions:
        if isinstance(params.custom_component_types_definitions, str):
            params.custom_component_types_definitions = [
                params.custom_component_types_definitions
            ]
        for ctd in params.custom_component_types_definitions:
            if target_directory != "./":
                # def_file = './%s' % ctd
                def_file = "%s/%s" % (os.path.dirname(os.path.abspath(__file__)), ctd)

                if (
                    param_overrides
                    and "custom_component_type_gate_overrides" in param_overrides
                    and param_overrides["custom_component_type_gate_overrides"]
                ):
                    root = etree.parse(def_file).getroot()

                    for k, v in param_overrides[
                        "custom_component_type_gate_overrides"
                    ].items():
                        channel_id = k.split("__")[0]
                        gate_id = k.split("__")[1]
                        gate_attr = k.split("__")[2]

                        for c1 in root:
                            if (
                                c1.tag != "ionChannel"
                                and c1.attrib.get("id") != channel_id
                            ):
                                continue
                            for c2 in c1:
                                if (
                                    c2.tag != "gateHHtauInf"
                                    and c2.attrib.get("id") != gate_id
                                ):
                                    continue
                                for child in c2:
                                    if child.attrib.get(gate_attr):
                                        child.set(gate_attr, v)
                    etree.ElementTree(root).write(
                        os.path.join(target_directory, ctd), pretty_print=True
                    )
                else:
                    shutil.copy(def_file, target_directory)

            if target_directory == "./" and not os.path.isfile(ctd):
                ctd = "%s/%s" % (os.path.dirname(__file__), ctd)

            lems_info["includes"].append(ctd)
            nml_doc.includes.append(IncludeType(href=ctd))

    import c302.backers

    cells_vs_name = c302.backers.get_adopted_cell_names()

    count = 0

    try:
        with Bundle("openworm/owmeta-data", version=6) as bnd:
            all_neuron_info, all_muscle_info = _get_cell_info(bnd, set(cell_names))
    except Exception as e:
        print_('Unable to open "openworm/owmeta-data" bundle: %s' % e)
        all_neuron_info, all_muscle_info = _get_cell_info(None, set(cell_names))

    for cell in cell_names:
        if cells is None or cell in cells:
            inst = Instance(id="0")

            if not params.is_level_D():
                # build a Population data structure out of the cell name
                pop0 = Population(
                    id=cell,
                    component=params.generic_neuron_cell.id,
                    type="populationList",
                    size="1",
                )
                cell_id = params.generic_neuron_cell.id
            else:
                # build a Population data structure out of the cell name
                pop0 = Population(
                    id=cell, component=cell, type="populationList", size="1"
                )
                cell_id = cell

            # neuron, neuron.type(), neuron.receptor(), neuron.neurotransmitter(), short, color
            if all_neuron_info is not None:
                pop0.properties.append(Property("color", all_neuron_info[cell][5]))
                types = sorted(all_neuron_info[cell][1])
                pop0.properties.append(Property("type", str("; ".join(types))))
                recps = sorted(all_neuron_info[cell][2])
                pop0.properties.append(Property("receptor", str("; ".join(recps))))
                nt_sorted = sorted(all_neuron_info[cell][3])
                pop0.properties.append(
                    Property("neurotransmitter", str("; ".join(nt_sorted)))
                )

            pop0.instances.append(inst)

            # put that Population into the Network data structure from above
            net.populations.append(pop0)

            if cell in cells_vs_name:
                p = Property(
                    tag="OpenWormBackerAssignedName", value=cells_vs_name[cell]
                )
                pop0.properties.append(p)

            # also use the cell name to grab the morphology file, as a NeuroML data structure
            #  into the 'all_cells' dict
            cell_file_path = (
                root_dir + "/../" if test else root_dir + "/"
            )  # if running test
            cell_file = cell_file_path + "NeuroML2/%s.cell.nml" % cell
            doc = loaders.NeuroMLLoader.load(cell_file)
            all_cells[cell] = doc.cells[0]

            if params.is_level_D():
                new_cell = params.create_neuron_cell(cell, doc.cells[0].morphology)

                nml_cell_doc = NeuroMLDocument(id=cell)
                nml_cell_doc.cells.append(new_cell)
                new_cell_file = "cells/" + cell + "_D.cell.nml"
                nml_file = target_directory + "/" + new_cell_file
                print_("Writing new cell to: %s" % os.path.realpath(nml_file))
                writers.NeuroMLWriter.write(nml_cell_doc, nml_file)

                nml_doc.includes.append(IncludeType(href=new_cell_file))
                lems_info["includes"].append(new_cell_file)

                inst.location = Location(0, 0, 0)
            else:
                location = doc.cells[0].morphology.segments[0].proximal

                inst.location = Location(
                    float(location.x), float(location.y), float(location.z)
                )

            if verbose:
                print_(
                    "Loaded morphology: %s; id: %s; placing at location: (%s, %s, %s)"
                    % (
                        os.path.realpath(cell_file),
                        all_cells[cell].id,
                        inst.location.x,
                        inst.location.y,
                        inst.location.z,
                    )
                )

            if cells_to_stimulate is None or cell in cells_to_stimulate:
                target = "../%s/0/%s" % (pop0.id, cell_id)

                input_list = InputList(
                    id="Input_%s_%s" % (cell, params.offset_current.id),
                    component=params.offset_current.id,
                    populations="%s" % cell,
                )

                i0 = Input(id=0, target=target, destination="synapses")

                if params.is_level_D():
                    i0.segment_id = 0

                input_list.input.append(i0)

                net.input_lists.append(input_list)

            if cells_to_plot is None or cell in cells_to_plot:
                plot = {}

                plot["cell"] = cell
                plot["colour"] = get_random_colour_hex()
                plot["quantity"] = "%s/0/%s/v" % (cell, cell_id)
                lems_info["plots"].append(plot)

                if params.is_level_B():
                    plot = {}

                    plot["cell"] = cell
                    plot["colour"] = get_random_colour_hex()
                    plot["quantity"] = "%s/0/%s/activity" % (cell, cell_id)
                    lems_info["activity_plots"].append(plot)

                if is_cond_based_cell(params):
                    plot = {}

                    plot["cell"] = cell
                    plot["colour"] = get_random_colour_hex()
                    plot["quantity"] = "%s/0/%s/caConc" % (cell, cell_id)
                    lems_info["activity_plots"].append(plot)

            save = {}
            save["cell"] = cell
            save["quantity"] = "%s/0/%s/v" % (cell, cell_id)
            lems_info["to_save"].append(save)

            if params.is_level_B():
                save = {}
                save["cell"] = cell
                save["quantity"] = "%s/0/%s/activity" % (cell, cell_id)
                lems_info["activity_to_save"].append(save)
            if is_cond_based_cell(params):
                save = {}
                save["cell"] = cell
                save["quantity"] = "%s/0/%s/caConc" % (cell, cell_id)
                lems_info["activity_to_save"].append(save)

            lems_info["cells"].append(cell)

            count += 1

    if verbose:
        print_("Finished loading %i cells" % count)

    mneurons, all_muscles, muscle_conns = get_cell_muscle_names_and_connection(
        data_reader
    )

    if muscles_to_include is None or muscles_to_include is True:
        muscles_to_include = all_muscles
    elif muscles_to_include is False:
        muscles_to_include = []

    for m in muscles_to_include:
        if m not in all_muscles:
            raise Exception("%s is not among the known muscles" % m)

    if len(muscles_to_include) > 0:
        muscle_count = 0
        for muscle in muscles_to_include:
            inst = Instance(id="0")

            # build a Population data structure out of the cell name
            pop0 = Population(
                id=muscle,
                component=params.generic_muscle_cell.id,
                type="populationList",
                size="1",
            )
            pop0.properties.append(Property("color", "0 .6 0"))
            pop0.instances.append(inst)

            # put that Population into the Network data structure from above
            net.populations.append(pop0)

            if muscle in cells_vs_name:
                # No muscles adopted yet, but just in case they are in future...
                p = Property(
                    tag="OpenWormBackerAssignedName", value=cells_vs_name[muscle]
                )
                pop0.properties.append(p)

            x, y, z = get_muscle_position(muscle, data_reader)
            # print_('Positioning muscle: %s at (%s,%s,%s)'%(muscle,x,y,z))
            inst.location = Location(x, y, z)

            # target = "%s/0/%s"%(pop0.id, params.generic_muscle_cell.id) # unused

            plot = {}

            plot["cell"] = muscle
            plot["colour"] = get_random_colour_hex()
            plot["quantity"] = "%s/0/%s/v" % (muscle, params.generic_muscle_cell.id)
            lems_info["muscle_plots"].append(plot)

            if params.generic_muscle_cell.__class__.__name__ == "IafActivityCell":
                plot = {}

                plot["cell"] = muscle
                plot["colour"] = get_random_colour_hex()
                plot["quantity"] = "%s/0/%s/activity" % (
                    muscle,
                    params.generic_muscle_cell.id,
                )
                lems_info["muscle_activity_plots"].append(plot)

            if params.generic_muscle_cell.__class__.__name__ == "Cell":
                plot = {}

                plot["cell"] = muscle
                plot["colour"] = get_random_colour_hex()
                plot["quantity"] = "%s/0/%s/caConc" % (
                    muscle,
                    params.generic_muscle_cell.id,
                )
                lems_info["muscle_activity_plots"].append(plot)

            save = {}
            save["cell"] = muscle
            save["quantity"] = "%s/0/%s/v" % (muscle, params.generic_muscle_cell.id)
            lems_info["muscles_to_save"].append(save)

            if params.generic_muscle_cell.__class__.__name__ == "IafActivityCell":
                save = {}
                save["cell"] = muscle
                save["quantity"] = "%s/0/%s/activity" % (
                    muscle,
                    params.generic_muscle_cell.id,
                )
                lems_info["muscles_activity_to_save"].append(save)
            if params.generic_muscle_cell.__class__.__name__ == "Cell":
                save = {}
                save["cell"] = muscle
                save["quantity"] = "%s/0/%s/caConc" % (
                    muscle,
                    params.generic_muscle_cell.id,
                )
                lems_info["muscles_activity_to_save"].append(save)

            lems_info["muscles"].append(muscle)

            muscle_count += 1

            if cells_to_stimulate is not None and muscle in cells_to_stimulate:
                target = "../%s/0/%s" % (pop0.id, params.generic_muscle_cell.id)

                input_list = InputList(
                    id="Input_%s_%s" % (muscle, params.offset_current.id),
                    component=params.offset_current.id,
                    populations="%s" % pop0.id,
                )

                i0 = Input(id=0, target=target, destination="synapses")

                if params.is_level_D():
                    i0.segment_id = 0

                input_list.input.append(i0)

                net.input_lists.append(input_list)

        if verbose:
            print_("Finished creating %i muscles" % muscle_count)

    existing_synapses = {}

    for conn in conns:
        if conn.pre_cell in lems_info["cells"] and conn.post_cell in lems_info["cells"]:
            # take information about each connection and package it into a
            # NeuroML Projection data structure
            proj_id = get_projection_id(
                conn.pre_cell, conn.post_cell, conn.synclass, conn.syntype
            )
            conn_shorthand = "%s-%s" % (conn.pre_cell, conn.post_cell)

            elect_conn = False
            analog_conn = False

            conn_type = "neuron_to_neuron"
            conn_pol = "exc"

            orig_pol = "exc"

            if "GABA" in conn.synclass:
                conn_pol = "inh"
                orig_pol = "inh"
            if "_GJ" in conn.synclass:
                conn_pol = "elec"
                elect_conn = params.is_elec_conn(params.neuron_to_neuron_elec_syn)
                conn_shorthand = "%s-%s_GJ" % (conn.pre_cell, conn.post_cell)

            for key in regex_param_overrides.keys():
                if key == "mirrored_elec_conn_params":
                    continue

                pattern = key.split("$")[0] + "$"
                pattern = pattern.replace("_to_", "-")

                if re.match(pattern, conn_shorthand):
                    new_param = conn_shorthand.replace("-", "_to_") + key.split("$")[1]
                    new_param = new_param.replace("_GJ", "")
                    new_param_v = regex_param_overrides[key]

                    if new_param in param_overrides:
                        continue
                    # add regex param unless there is a specific param
                    set_param(params, new_param, new_param_v)

            if "mirrored_elec_conn_params" in regex_param_overrides:
                for k, v in regex_param_overrides["mirrored_elec_conn_params"].items():
                    pattern = k.split("$")[0] + "$"
                    pattern = pattern.replace("_to_", "-")

                    if re.match(pattern, conn_shorthand):
                        new_param = (
                            conn_shorthand.replace("-", "_to_") + k.split("$")[1]
                        )
                        new_param = new_param.replace("_GJ", "")
                        new_param_mirrored = (
                            conn.post_cell
                            + "_"
                            + new_param.split("_")[1]
                            + "_"
                            + conn.pre_cell
                            + "_"
                            + "_".join(new_param.split("_")[3:])
                        )
                        new_param_v = v

                        if (
                            new_param in param_overrides["mirrored_elec_conn_params"]
                            or new_param_mirrored
                            in param_overrides["mirrored_elec_conn_params"]
                        ):
                            continue
                        # add regex param unless there is a specific param
                        mirror_param(params, new_param, new_param_v)

            if conns_to_include and conn_shorthand not in conns_to_include:
                # conn_shorthand not in conns_to_include. if there is a regex in conns_to_include which matches the current conn_shorthand -> include
                if not elem_in_coll_matches_conn(conns_to_include, conn_shorthand):
                    continue
            if conns_to_exclude:
                if conn_shorthand in conns_to_exclude:
                    continue
                # conn_shorthand not in conns_to_exclude. if there is a regex in conns_to_exclude which matches the current conn_shorthand -> exclude
                if elem_in_coll_matches_conn(conns_to_exclude, conn_shorthand):
                    continue

            syn0 = params.get_syn(conn.pre_cell, conn.post_cell, conn_type, conn_pol)

            if print_connections:
                print_(
                    conn_shorthand
                    + " "
                    + str(conn.number)
                    + " "
                    + orig_pol
                    + " "
                    + conn.synclass
                    + " "
                    + syn0.id
                )

            polarity = None
            if conn_polarity_override:
                # if elem_in_coll_matches_conn(conn_polarity_override.keys(), conn_shorthand):
                for conn_pol in conn_polarity_override.keys():
                    if conn_pol == conn_shorthand:
                        polarity = conn_polarity_override[conn_shorthand]
                        break
                    elif regex_match(conn_pol, conn_shorthand):
                        polarity = conn_polarity_override[conn_pol]
                        break
            if polarity and not elect_conn:
                syn0 = params.get_syn(
                    conn.pre_cell, conn.post_cell, conn_type, polarity
                )
                if verbose and polarity != orig_pol:
                    print_(
                        ">> Changing polarity of connection %s -> %s: was: %s, becomes %s "
                        % (conn.pre_cell, conn.post_cell, orig_pol, polarity)
                    )

            if params.is_analog_conn(syn0):
                analog_conn = True
                if len(nml_doc.silent_synapses) == 0:
                    nml_doc.silent_synapses.append(SilentSynapse(id="silent"))

            number_syns = conn.number

            if params.get_bioparameter(
                "global_connectivity_power_scaling", warn_if_missing=False
            ):
                scale = params.get_bioparameter("global_connectivity_power_scaling").x()
                # print("Scaling by %s"%scale)
                number_syns = math.pow(number_syns, scale)

            if conn_number_override:
                # number_syns = conn_number_override[conn_shorthand]

                for conn_num_override in conn_number_override.keys():
                    if conn_num_override == conn_shorthand:
                        number_syns = conn_number_override[conn_shorthand]
                        break
                    elif regex_match(conn_num_override, conn_shorthand):
                        number_syns = conn_number_override[conn_num_override]
                        break
            if conn_number_scaling:
                # number_syns = conn.number*conn_number_scaling[conn_shorthand]

                for conn_num_scale in conn_number_scaling.keys():
                    if conn_num_scale == conn_shorthand:
                        number_syns = conn.number * conn_number_scaling[conn_shorthand]
                        break
                    elif regex_match(conn_num_scale, conn_shorthand):
                        number_syns = conn.number * conn_number_scaling[conn_num_scale]
                        break
            """
            else:
                print(conn_shorthand)
                print(conn_number_override)
                print(conn_number_scaling)"""
            """if polarity:
                print "%s %s num:%s" % (conn_shorthand, polarity, number_syns)
            elif elect_conn:
                print "%s num:%s" % (conn_shorthand, number_syns)
            else:
                print "%s %s num:%s" % (conn_shorthand, orig_pol, number_syns)"""

            if number_syns != conn.number:
                if analog_conn or elect_conn:
                    magnitude, unit = c302.bioparameters.split_neuroml_quantity(
                        syn0.conductance
                    )
                else:
                    magnitude, unit = c302.bioparameters.split_neuroml_quantity(
                        syn0.gbase
                    )
                cond0 = "%s%s" % (magnitude * conn.number, unit)
                cond1 = "%s%s" % (
                    get_str_from_expnotation(magnitude * number_syns),
                    unit,
                )
                gj = "" if not elect_conn else " GapJunction"
                if verbose:
                    print_(
                        ">> Changing number of effective synapses connection %s -> %s%s: was: %s (total cond: %s), becomes %s (total cond: %s)"
                        % (
                            conn.pre_cell,
                            conn.post_cell,
                            gj,
                            conn.number,
                            cond0,
                            number_syns,
                            cond1,
                        )
                    )

            # print "######## %s-%s %s %s" % (conn.pre_cell, conn.post_cell, conn.synclass, number_syns)
            # known_motor_prefixes = ["VA"]
            # if conn.pre_cell.startswith(tuple(known_motor_prefixes)) or conn.post_cell.startswith(tuple(known_motor_prefixes)):
            #    print "######### %s-%s %s %s" % (conn.pre_cell, conn.post_cell, number_syns, conn.synclass)

            syn_new = params.create_n_connection_synapse(
                syn0, number_syns, nml_doc, existing_synapses
            )

            if elect_conn:
                proj0 = ElectricalProjection(
                    id=proj_id,
                    presynaptic_population=conn.pre_cell,
                    postsynaptic_population=conn.post_cell,
                )

                net.electrical_projections.append(proj0)

                pre_cell_id = get_cell_id_string(conn.pre_cell, params)
                post_cell_id = get_cell_id_string(conn.post_cell, params)

                # print_("Conn %s -> %s"%(pre_cell_id,post_cell_id))

                # Add a Connection with the closest locations
                conn0 = ElectricalConnectionInstanceW(
                    id="0",
                    pre_cell=pre_cell_id,
                    post_cell=post_cell_id,
                    synapse=syn_new.id,
                    weight=number_syns,
                )

                proj0.electrical_connection_instance_ws.append(conn0)

            elif analog_conn:
                proj0 = ContinuousProjection(
                    id=proj_id,
                    presynaptic_population=conn.pre_cell,
                    postsynaptic_population=conn.post_cell,
                )

                net.continuous_projections.append(proj0)

                pre_cell_id = get_cell_id_string(conn.pre_cell, params)
                post_cell_id = get_cell_id_string(conn.post_cell, params)

                conn0 = ContinuousConnectionInstanceW(
                    id="0",
                    pre_cell=pre_cell_id,
                    post_cell=post_cell_id,
                    pre_component="silent",
                    post_component=syn_new.id,
                    weight=number_syns,
                )

                proj0.continuous_connection_instance_ws.append(conn0)

            else:
                proj0 = Projection(
                    id=proj_id,
                    presynaptic_population=conn.pre_cell,
                    postsynaptic_population=conn.post_cell,
                    synapse=syn_new.id,
                )

                net.projections.append(proj0)

                pre_cell_id = get_cell_id_string(conn.pre_cell, params)
                post_cell_id = get_cell_id_string(conn.post_cell, params)

                conn0 = ConnectionWD(
                    id="0",
                    pre_cell_id=pre_cell_id,
                    post_cell_id=post_cell_id,
                    weight=number_syns,
                    delay="0ms",
                )

                proj0.connection_wds.append(conn0)

    if len(muscles_to_include) > 0:
        for conn in muscle_conns:
            if conn.post_cell not in muscles_to_include:
                continue
            if (
                conn.pre_cell not in lems_info["cells"]
                and conn.pre_cell not in muscles_to_include
            ):
                continue

            # take information about each connection and package it into a
            # NeuroML Projection data structure
            proj_id = get_projection_id(
                conn.pre_cell, conn.post_cell, conn.synclass, conn.syntype
            )
            conn_shorthand = "%s-%s" % (conn.pre_cell, conn.post_cell)

            elect_conn = False
            analog_conn = False

            conn_type = "neuron_to_muscle"
            if conn.pre_cell in muscles_to_include:
                conn_type = "muscle_to_muscle"
            # if conn.post_cell in lems_info['cells']:
            #    conn_type = "muscle_to_neuron"
            conn_pol = "exc"
            orig_pol = "exc"

            if "GABA" in conn.synclass:
                conn_pol = "inh"
                orig_pol = "inh"

            if "_GJ" in conn.synclass:
                conn_pol = "elec"
                orig_pol = "elec"
                elect_conn = params.is_elec_conn(params.neuron_to_neuron_elec_syn)
                conn_shorthand = "%s-%s_GJ" % (conn.pre_cell, conn.post_cell)

            for key in regex_param_overrides.keys():
                if key == "mirrored_elec_conn_params":
                    continue

                pattern = key.split("$")[0] + "$"
                pattern = pattern.replace("_to_", "-")

                if re.match(pattern, conn_shorthand):
                    new_param = conn_shorthand.replace("-", "_to_") + key.split("$")[1]
                    new_param = new_param.replace("_GJ", "")
                    new_param_v = regex_param_overrides[key]

                    if new_param in param_overrides:
                        continue
                    # add regex param unless there is a specific param
                    set_param(params, new_param, new_param_v)

            if "mirrored_elec_conn_params" in regex_param_overrides:
                for k, v in regex_param_overrides["mirrored_elec_conn_params"].items():
                    pattern = k.split("$")[0] + "$"
                    pattern = pattern.replace("_to_", "-")

                    if re.match(pattern, conn_shorthand):
                        new_param = (
                            conn_shorthand.replace("-", "_to_") + k.split("$")[1]
                        )
                        new_param = new_param.replace("_GJ", "")
                        new_param_mirrored = (
                            conn.post_cell
                            + "_"
                            + new_param.split("_")[1]
                            + "_"
                            + conn.pre_cell
                            + "_"
                            + "_".join(new_param.split("_")[3:])
                        )
                        new_param_v = v

                        if (
                            new_param in param_overrides["mirrored_elec_conn_params"]
                            or new_param_mirrored
                            in param_overrides["mirrored_elec_conn_params"]
                        ):
                            continue
                        # add regex param unless there is a specific param
                        mirror_param(params, new_param, new_param_v)

            if conns_to_include and conn_shorthand not in conns_to_include:
                # conn_shorthand not in conns_to_include. if there is a regex in conns_to_include which matches the current conn_shorthand -> include
                if not elem_in_coll_matches_conn(conns_to_include, conn_shorthand):
                    continue
            if conns_to_exclude:
                if conn_shorthand in conns_to_exclude:
                    continue
                # conn_shorthand not in conns_to_exclude. if there is a regex in conns_to_exclude which matches the current conn_shorthand -> exclude
                if elem_in_coll_matches_conn(conns_to_exclude, conn_shorthand):
                    continue

            syn0 = params.get_syn(conn.pre_cell, conn.post_cell, conn_type, conn_pol)

            if print_connections:
                print_(
                    conn_shorthand
                    + " "
                    + str(conn.number)
                    + " "
                    + orig_pol
                    + " "
                    + conn.synclass
                )

            polarity = None
            if conn_polarity_override:
                for conn_pol in conn_polarity_override.keys():
                    if conn_pol == conn_shorthand:
                        polarity = conn_polarity_override[conn_pol]
                        break
                    elif regex_match(conn_pol, conn_shorthand):
                        polarity = conn_polarity_override[conn_pol]
                        break
            if polarity and not elect_conn:
                syn0 = params.get_syn(
                    conn.pre_cell, conn.post_cell, conn_type, polarity
                )
                if verbose and polarity != orig_pol:
                    print_(
                        ">> Changing polarity of connection %s -> %s: was: %s, becomes %s "
                        % (conn.pre_cell, conn.post_cell, orig_pol, polarity)
                    )

            if params.is_analog_conn(syn0):
                analog_conn = True
                if len(nml_doc.silent_synapses) == 0:
                    nml_doc.silent_synapses.append(SilentSynapse(id="silent"))

            number_syns = conn.number

            if params.get_bioparameter(
                "global_connectivity_power_scaling", warn_if_missing=False
            ):
                scale = params.get_bioparameter("global_connectivity_power_scaling").x()
                # print("Scaling by %s"%scale)
                number_syns = math.pow(number_syns, scale)

            if conn_number_override:
                for conn_num_override in conn_number_override.keys():
                    if conn_num_override == conn_shorthand:
                        number_syns = conn_number_override[conn_shorthand]
                        break
                    elif regex_match(conn_num_override, conn_shorthand):
                        number_syns = conn_number_override[conn_num_override]
                        break
            if conn_number_scaling:
                for conn_num_scale in conn_number_scaling.keys():
                    if conn_num_scale == conn_shorthand:
                        number_syns = conn.number * conn_number_scaling[conn_shorthand]
                        break
                    elif regex_match(conn_num_scale, conn_shorthand):
                        number_syns = conn.number * conn_number_scaling[conn_num_scale]
                        break
            """
            else:
                print(conn_shorthand)
                print(conn_number_override)
                print(conn_number_scaling)"""
            """if polarity:
                print "%s %s num:%s" % (conn_shorthand, polarity, number_syns)
            elif elect_conn:
                print "%s num:%s" % (conn_shorthand, number_syns)
            else:
                print "%s %s num:%s" % (conn_shorthand, orig_pol, number_syns)"""

            if number_syns != conn.number:
                if analog_conn or elect_conn:
                    magnitude, unit = c302.bioparameters.split_neuroml_quantity(
                        syn0.conductance
                    )
                else:
                    magnitude, unit = c302.bioparameters.split_neuroml_quantity(
                        syn0.gbase
                    )
                cond0 = "%s%s" % (magnitude * conn.number, unit)
                cond1 = "%s%s" % (
                    get_str_from_expnotation(magnitude * number_syns),
                    unit,
                )
                gj = "" if not elect_conn else " GapJunction"
                if verbose:
                    print_(
                        ">> Changing number of effective synapses connection %s -> %s%s: was: %s (total cond: %s), becomes %s (total cond: %s)"
                        % (
                            conn.pre_cell,
                            conn.post_cell,
                            gj,
                            conn.number,
                            cond0,
                            number_syns,
                            cond1,
                        )
                    )

            syn_new = params.create_n_connection_synapse(
                syn0, number_syns, nml_doc, existing_synapses
            )

            if elect_conn:
                proj0 = ElectricalProjection(
                    id=proj_id,
                    presynaptic_population=conn.pre_cell,
                    postsynaptic_population=conn.post_cell,
                )

                net.electrical_projections.append(proj0)

                pre_cell_id = get_cell_id_string(conn.pre_cell, params)
                post_cell_id = get_cell_id_string(conn.post_cell, params, muscle=True)

                # print_("Conn %s -> %s"%(pre_cell_id,post_cell_id))

                # Add a Connection with the closest locations
                conn0 = ElectricalConnectionInstanceW(
                    id="0",
                    pre_cell=pre_cell_id,
                    post_cell=post_cell_id,
                    synapse=syn_new.id,
                    weight=number_syns,
                )

                proj0.electrical_connection_instance_ws.append(conn0)

            elif analog_conn:
                proj0 = ContinuousProjection(
                    id=proj_id,
                    presynaptic_population=conn.pre_cell,
                    postsynaptic_population=conn.post_cell,
                )

                net.continuous_projections.append(proj0)

                pre_cell_id = get_cell_id_string(conn.pre_cell, params)
                post_cell_id = get_cell_id_string(conn.post_cell, params, muscle=True)

                conn0 = ContinuousConnectionInstanceW(
                    id="0",
                    pre_cell=pre_cell_id,
                    post_cell=post_cell_id,
                    pre_component="silent",
                    post_component=syn_new.id,
                    weight=number_syns,
                )

                proj0.continuous_connection_instance_ws.append(conn0)

            else:
                proj0 = Projection(
                    id=proj_id,
                    presynaptic_population=conn.pre_cell,
                    postsynaptic_population=conn.post_cell,
                    synapse=syn_new.id,
                )

                net.projections.append(proj0)

                # Add a Connection with the closest locations

                pre_cell_id = get_cell_id_string(conn.pre_cell, params)
                post_cell_id = get_cell_id_string(conn.post_cell, params, muscle=True)

                conn0 = Connection(
                    id="0", pre_cell_id=pre_cell_id, post_cell_id=post_cell_id
                )

                proj0.connections.append(conn0)

    if param_overrides and param_overrides.keys():
        info_new = info_settings + "\n%s\n" % (params.bioparameter_info("    "))
        nml_doc.notes = info_new
        lems_info["comment"] = info_new

    # import pprint
    # pprint.pprint(lems_info)
    template_path = root_dir + "/../" if test else root_dir + "/"  # if running test
    write_to_file(
        nml_doc,
        lems_info,
        net_id,
        template_path,
        validate=validate,
        verbose=verbose,
        target_directory=target_directory,
    )

    return nml_doc


"""
    Input:    string of form ["AVAL","AVBL"]
    returns:  ["AVAL", "AVBL"]
"""


def parse_list_arg(list_arg):
    if list_arg is None:
        return None
    if list_arg == []:
        return []
    entries = list_arg[1:-1].split(",")
    ret = [e for e in entries]
    print_("Command line argument %s parsed as: %s" % (list_arg, ret))
    return ret


"""
    Input:    string of form ["ADAL-AIBL":2.5,"I1L-I1R":0.5]
    returns:  {}
"""


def parse_dict_arg(dict_arg):
    if dict_arg is None or dict_arg == "None":
        return None
    ret = {}
    entries = str(dict_arg[1:-1]).split(",")
    for e in entries:
        try:
            ret[e.split(":")[0]] = float(e.split(":")[1])  # {'AVAL-AVAR':1}
        except ValueError:
            ret[e.split(":")[0]] = str(e.split(":")[1])  # {'AVAL-AVAR':'inh'}
    print_("Command line argument %s parsed as: %s" % (dict_arg, ret))
    return ret


def main():
    import importlib

    print("Starting c302 v%s..." % __version__)
    args = process_args()

    ParameterisedModel = getattr(
        importlib.import_module("c302.%s" % args.parameters),
        "ParameterisedModel",
    )
    params = ParameterisedModel()
    generate(
        args.reference,
        params,
        data_reader=args.datareader,
        cells=parse_list_arg(args.cells),
        cells_to_plot=parse_list_arg(args.cellstoplot),
        cells_to_stimulate=parse_list_arg(args.cellstostimulate),
        conn_polarity_override=parse_dict_arg(args.connpolarityoverride),
        conn_number_override=parse_dict_arg(args.connnumberoverride),
        conn_number_scaling=parse_dict_arg(args.connnumberscaling),
        muscles_to_include=parse_list_arg(args.musclestoinclude),
        param_overrides=parse_dict_arg(args.paramoverride),
        duration=args.duration,
        dt=args.dt,
        vmin=args.vmin,
        vmax=args.vmax,
    )


if __name__ == "__main__":
    import sys

    if "-cache" in sys.argv:
        print("Starting c302...")
        from c302.ConnectomeReader import PREFERRED_MUSCLE_NAMES
        from c302.ConnectomeReader import PREFERRED_NEURON_NAMES

        all_info = {"neuron_info": {}, "muscle_info": {}}

        from owmeta_core.bundle import Bundle

        from owmeta_core import __version__ as owc_version
        from owmeta import __version__ as owmeta_version

        ver_info = "owmeta v%s (owmeta core v%s)" % (owmeta_version, owc_version)
        all_info["comment"] = "Information exported from " + ver_info

        with Bundle("openworm/owmeta-data", version=6) as bnd:
            for n in PREFERRED_NEURON_NAMES:
                ani, _all_muscle_info = _get_cell_info(bnd, [n])

                print("  > %s; %s" % (ani[n], _all_muscle_info))
                all_info["neuron_info"][n] = (
                    str(ani[n][0]),
                    sorted(list(ani[n][1])),
                    sorted(list(ani[n][2])),
                    sorted(list(ani[n][3])),
                    ani[n][4],
                    ani[n][5],
                )

            for n in PREFERRED_MUSCLE_NAMES:
                _all_neuron_info, ami = _get_cell_info(bnd, [n])

                if not n == "MVULVA" and not n == "MANAL":
                    ow_name = n[1:] if n[3] != "0" else "%s%s" % (n[1:3], n[-1])
                    print("  > %s (%s): %s; %s" % (n, ow_name, _all_neuron_info, ami))

                    all_info["muscle_info"][n] = (
                        str(ami[ow_name][0]),
                        sorted(list(ami[ow_name][1])),
                        sorted(list(ami[ow_name][2])),
                        sorted(list(ami[ow_name][3])),
                        ami[ow_name][4],
                        ami[ow_name][5],
                    )

        with open(OWMETA_CACHED_DATA_FILE, "w") as fp:
            json.dump(all_info, fp, sort_keys=True, indent=4)

    else:
        main()



# Temporary class to allow this to be used in comparison notebook.
# Should be tidied up.

from c302.WhiteDataReader import White_A

from c302.ConnectomeReader import analyse_connections

read_data = White_A.read_data
read_muscle_data = White_A.read_muscle_data


def main1():
    cells, neuron_conns = read_data(include_nonconnected_cells=True)
    neurons2muscles, muscles, muscle_conns = read_muscle_data()
    analyse_connections(cells, neuron_conns, neurons2muscles, muscles, muscle_conns)


if __name__ == "__main__":
    main1()



import sys
import os
import importlib

sys.path.insert(0, os.path.abspath("."))

import c302

import neuroml.writers as writers

range_incl = lambda start, end: range(start, end + 1)


def setup(
    parameter_set,
    generate=False,
    duration=2000,
    dt=0.05,
    target_directory="examples",
    data_reader=c302.DEFAULT_DATA_READER,
    param_overrides={},
    verbose=True,
    config_param_overrides={},
):
    ParameterisedModel = getattr(
        importlib.import_module("c302.parameters_%s" % parameter_set),
        "ParameterisedModel",
    )
    params = ParameterisedModel()

    params.set_bioparameter(
        "unphysiological_offset_current", "0pA", "Testing TapWithdrawal", "0"
    )
    params.set_bioparameter(
        "unphysiological_offset_current_del", "0 ms", "Testing TapWithdrawal", "0"
    )
    params.set_bioparameter(
        "unphysiological_offset_current_dur", "2000 ms", "Testing TapWithdrawal", "0"
    )

    """
    VA_motors = ["VA%s" % c for c in range_incl(1, 12)]
    VB_motors = ["VB%s" % c for c in range_incl(1, 11)]
    DA_motors = ["DA%s" % c for c in range_incl(1, 9)]
    DB_motors = ["DB%s" % c for c in range_incl(1, 7)]
    DD_motors = ["DD%s" % c for c in range_incl(1, 6)]
    VD_motors = ["VD%s" % c for c in range_incl(1, 13)]
    AS_motors = ["AS%s" % c for c in range_incl(1, 11)]"""

    cells = []

    muscles_to_include = True

    if config_param_overrides.has_key("muscles_to_include"):
        muscles_to_include = config_param_overrides["muscles_to_include"]

    cells_to_stimulate = []

    cells_to_plot = list(cells)
    reference = "c302_%s_MuscleTest" % parameter_set

    conns_to_include = []
    if config_param_overrides.has_key("conns_to_include"):
        conns_to_include = config_param_overrides["conns_to_include"]

    conns_to_exclude = ["^.+-.+$"]
    if config_param_overrides.has_key("conns_to_exclude"):
        conns_to_exclude = config_param_overrides["conns_to_exclude"]

    conn_polarity_override = {}
    if config_param_overrides.has_key("conn_polarity_override"):
        conn_polarity_override.update(config_param_overrides["conn_polarity_override"])

    conn_number_override = {}
    if config_param_overrides.has_key("conn_number_override"):
        conn_number_override.update(config_param_overrides["conn_number_override"])

    param_overrides = {
        "ca_conc_decay_time_muscle": "60 ms",
        "ca_conc_rho_muscle": "0.002138919 mol_per_m_per_A_per_s",
    }

    # end = "%sms" % (int(duration) - 100)

    input_list = []

    # input_list.append(('MDL02', '0ms', '250ms', '3pA'))
    # input_list.append(('MDL03', '0ms', '250ms', '3pA'))
    # input_list.append(('MDR02', '0ms', '250ms', '3pA'))
    # input_list.append(('MDR03', '0ms', '250ms', '3pA'))

    input_list.append(("MVR10", "0ms", "250ms", "1pA"))
    input_list.append(("MVR11", "0ms", "250ms", "2pA"))
    input_list.append(("MVR12", "0ms", "250ms", "3pA"))
    input_list.append(("MVR13", "0ms", "250ms", "3pA"))
    input_list.append(("MVR14", "0ms", "250ms", "2pA"))
    input_list.append(("MVR15", "0ms", "250ms", "1pA"))

    input_list.append(("MVL10", "0ms", "250ms", "1pA"))
    input_list.append(("MVL11", "0ms", "250ms", "2pA"))
    input_list.append(("MVL12", "0ms", "250ms", "3pA"))
    input_list.append(("MVL13", "0ms", "250ms", "3pA"))
    input_list.append(("MVL14", "0ms", "250ms", "2pA"))
    input_list.append(("MVL15", "0ms", "250ms", "1pA"))

    input_list.append(("MDL21", "0ms", "250ms", "3pA"))
    input_list.append(("MDL22", "0ms", "250ms", "3pA"))
    input_list.append(("MDR21", "0ms", "250ms", "3pA"))
    input_list.append(("MDR22", "0ms", "250ms", "3pA"))

    for stim_num in range(5):
        for muscle_num in range(24):
            mdlx = "MDL0%s" % (muscle_num + 1)
            mdrx = "MDR0%s" % (muscle_num + 1)

            mvlx = "MVL0%s" % (muscle_num + 1)
            mvrx = "MVR0%s" % (muscle_num + 1)

            if muscle_num >= 9:
                mdlx = "MDL%s" % (muscle_num + 1)
                mdrx = "MDR%s" % (muscle_num + 1)

                mvlx = "MVL%s" % (muscle_num + 1)
                if muscle_num != 23:
                    mvrx = "MVR%s" % (muscle_num + 1)

            startd = "%sms" % (stim_num * 1000 + muscle_num * 50)
            startv = "%sms" % ((stim_num * 1000 + 500) + muscle_num * 50)
            dur = "250ms"
            amp = "3pA"

            input_list.append((mdlx, startd, dur, amp))
            input_list.append((mdrx, startd, dur, amp))

            input_list.append((mvlx, startv, dur, amp))
            input_list.append((mvrx, startv, dur, amp))

    nml_doc = None
    if generate:
        nml_doc = c302.generate(
            reference,
            params,
            cells=cells,
            cells_to_plot=cells_to_plot,
            cells_to_stimulate=cells_to_stimulate,
            conns_to_include=conns_to_include,
            conns_to_exclude=conns_to_exclude,
            conn_polarity_override=conn_polarity_override,
            conn_number_override=conn_number_override,
            muscles_to_include=muscles_to_include,
            duration=duration,
            dt=dt,
            target_directory=target_directory,
            data_reader=data_reader,
            param_overrides=param_overrides,
            verbose=verbose,
        )

        if "input" in config_param_overrides:
            input_list = config_param_overrides["input"]

        for stim_input in input_list:
            cell, start, dur, current = stim_input
            c302.add_new_input(nml_doc, cell, start, dur, current, params)

        nml_file = target_directory + "/" + reference + ".net.nml"
        writers.NeuroMLWriter.write(
            nml_doc, nml_file
        )  # Write over network file written above...

        c302.print_("(Re)written network file to: " + nml_file)

    return cells, cells_to_stimulate, params, muscles_to_include, nml_doc


if __name__ == "__main__":
    parameter_set = sys.argv[1] if len(sys.argv) == 2 else "C2"
    setup_kwargs = dict()
    if len(sys.argv) == 3:
        setup_kwargs["data_reader"] = sys.argv[2]

    # setup(parameter_set, generate=True, data_reader=data_reader)



from decimal import Decimal

from neuroml import ExpTwoSynapse, GapJunction, GradedSynapse

"""
    Subject to much change & refactoring once owmeta is stable...
"""


def split_neuroml_quantity(quantity):
    i = len(quantity)
    while i > 0:
        magnitude = quantity[0:i].strip()
        unit = quantity[i:].strip()

        try:
            magnitude = float(magnitude)
            i = 0
        except ValueError:
            i -= 1
    return magnitude, unit


class BioParameter:
    def __init__(self, name, value, source, certainty):
        self.name = name
        self.value = value
        self.source = source
        self.certainty = certainty

    def __str__(self):
        return "BioParameter: %s = %s (SRC: %s, certainty %s)" % (
            self.name,
            self.value,
            self.source,
            self.certainty,
        )

    def __repr__(self):
        return self.__str__()

    def change_magnitude(self, magnitude):
        self.value = "%s %s" % (
            Decimal(magnitude),
            split_neuroml_quantity(self.value)[1],
        )

    def x(self):
        return split_neuroml_quantity(self.value)[0]


class ParameterisedModelPrototype(object):
    # bioparameters = []
    def __init__(self):
        self.bioparameters = []

    def print_(self, msg):
        pre = "c302      >>> "
        print("%s %s" % (pre, msg.replace("\n", "\n" + pre)))

    def add_bioparameter(self, name, value, source, certainty):
        found = False
        for bp in self.bioparameters:
            if bp.name == name:
                bp.value = value
                bp.source = source
                bp.certainty = certainty
                found = True
        if not found:
            bp = BioParameter(name, value, source, certainty)
            self.bioparameters.append(bp)

    def add_bioparameter_obj(self, bioparameter):
        for bp in self.bioparameters:
            if bp.name == bioparameter.name:
                self.bioparameters.remove(bp)

        self.bioparameters.append(bioparameter)

    def get_bioparameter(self, name, warn_if_missing=False):
        for bp in self.bioparameters:
            if bp.name == name:
                return bp
        if warn_if_missing:
            self.print_(
                "Cannot find bioparameter: %s; %i known: %s"
                % (
                    name,
                    len(self.bioparameters),
                    [bp.name for bp in self.bioparameters],
                )
            )
        return None

    def set_bioparameter(self, name, value, source, certainty):
        for bp in self.bioparameters:
            if bp.name == name:
                bp.value = value
                bp.source = source
                bp.certainty = certainty

    def bioparameter_info(self, indent=""):
        info = indent + "Known BioParameters:\n"
        for bp in sorted(self.bioparameters, key=lambda x: x.name):
            info += indent + indent + "%s\n" % bp
        return info


class NonNeuroMLCustomType:
    def __init__(self, id):
        self.id = id


class c302ModelPrototype(ParameterisedModelPrototype):
    def __init__(self):
        super(c302ModelPrototype, self).__init__()

        self.level = "Level not yet set"
        self.custom_component_types_definitions = None
        self.generic_neuron_cell = None
        self.generic_muscle_cell = None
        self.exc_syn = None
        self.inh_syn = None
        self.elec_syn = None
        self.offset_current = None
        self.concentration_model = None
        self.found_specific_param = False

    def is_level_A(self):
        return self.level.startswith("A")

    def is_level_B(self):
        return self.level.startswith("B")

    def is_level_C(self):
        return self.level.startswith("C")

    def is_level_C0(self):
        return self.level == "C0"

    def is_level_C2(self):
        return self.level == "C2"

    def is_level_D1(self):
        return self.level == "D1"

    def is_level_D(self):
        return self.level.startswith("D")

    def is_level_X(self):
        return self.level.startswith("X")

    def get_conn_param(
        self,
        pre_cell,
        post_cell,
        specific_conn_template,
        default_conn_template,
        param_name,
    ):
        param = self.get_bioparameter(
            specific_conn_template % (pre_cell, post_cell, param_name),
            warn_if_missing=False,
        )
        if param:
            self.found_specific_param = True
            return param.value
        def_param = self.get_bioparameter(default_conn_template % param_name)
        if not def_param:
            return None
        return def_param.value

    def get_syn(self, pre_cell, post_cell, type, pol):
        if pol == "elec":
            return self.get_elec_syn(pre_cell, post_cell, type)
        elif pol == "exc":
            return self.get_exc_syn(pre_cell, post_cell, type)
        elif pol == "inh":
            return self.get_inh_syn(pre_cell, post_cell, type)

    def create_n_connection_synapse(self, prototype_syn, n, nml_doc, existing_synapses):
        if prototype_syn.id in existing_synapses:
            return existing_synapses[prototype_syn.id]

        existing_synapses[prototype_syn.id] = prototype_syn
        if isinstance(prototype_syn, ExpTwoSynapse):
            nml_doc.exp_two_synapses.append(prototype_syn)
        elif isinstance(prototype_syn, GapJunction):
            nml_doc.gap_junctions.append(prototype_syn)
        elif isinstance(prototype_syn, GradedSynapse):
            nml_doc.graded_synapses.append(prototype_syn)
        elif isinstance(prototype_syn, NonNeuroMLCustomType):
            pass
        else:
            info = "%s conns; %s" % (n, existing_synapses)
            del existing_synapses[prototype_syn.id]
            raise Exception("Unknown synapse type: %s (%s)" % (prototype_syn.id, info))

        return prototype_syn

    def is_analog_conn(self, syn):
        return isinstance(syn, GradedSynapse) or isinstance(syn, NonNeuroMLCustomType)

    def is_elec_conn(self, syn):
        return isinstance(syn, GapJunction)
